import torch

import numpy as np
import torch.nn as nn

from typing import *
from cftool.misc import *
from cfdata.tabular import *
from cftool.ml import Metrics

from .loss import DDRLoss
from ..fcnn.core import FCNN
from ...misc.toolkit import *
from ...modules.blocks import *

tensors_type = Union[torch.Tensor, List[torch.Tensor], Dict[str, Union[torch.Tensor, tensor_dict_type]]]


class BasicInfo(NamedTuple):
    median: torch.Tensor
    median_detach: torch.Tensor
    feature_layers: List[torch.Tensor]
    anchor_batch: Union[torch.Tensor, None]
    quantile_batch: Union[torch.Tensor, None]


@FCNN.register("ddr")
class DDR(FCNN):
    def __init__(self,
                 config: Dict[str, Any],
                 tr_data: TabularData,
                 device: torch.device):
        if tr_data.task_type is not TaskTypes.REGRESSION:
            raise ValueError("DDR can only deal with regression problems")
        super().__init__(config, tr_data, device)
        self.__feature_params, self.__reg_params = [], []
        self._inject_median_params()
        self._init_cdf_layers()
        self._init_quantile_layers()
        self._init_parameters()
        self.q_metric = Metrics("quantile")
        labels = tr_data.processed.y
        self.y_min, self.y_max = labels.min(), labels.max()
        self.y_diff = self.y_max - self.y_min
        self._step_count = 0

    def _init_config(self,
                     tr_data: TabularData):
        # common mapping configs
        self._common_configs = self.config.setdefault("common_configs", {})
        self._common_configs.setdefault("pruner_config", None)
        self._common_configs.setdefault("batch_norm", False)
        self._common_configs.setdefault("dropout", False)
        # feature mappings
        self.feature_units = self.config.setdefault("hidden_units", [512, 512])
        mapping_configs = self.config.setdefault("mapping_configs", {})
        if isinstance(mapping_configs, dict):
            mapping_configs.setdefault("activation", "mish")
            mapping_configs.setdefault("init_method", "xavier")
            mapping_configs.setdefault("batch_norm", False)
            mapping_configs.setdefault("dropout", False)
            mapping_configs.setdefault("bias", False)
            mapping_configs = [mapping_configs] * self.num_feature_layers
        self.median_units = self.config.setdefault("median_units", [512])
        self._q_reg_activation = self.config.setdefault("quantile_reg_activation", "ReLU")
        self._reg_init = self.config.setdefault("regression_initialization", "xavier")
        median_mapping_configs = self.config.setdefault(
            "median_mapping_configs", shallow_copy_dict(self._common_configs))
        if isinstance(median_mapping_configs, dict):
            median_mapping_configs.setdefault("activation", self._q_reg_activation)
            median_mapping_configs.setdefault("init_method", self._reg_init)
            median_mapping_configs = [median_mapping_configs] * len(self.median_units)
        self.config["hidden_units"] = self.feature_units + self.median_units
        self.config["mapping_configs"] = mapping_configs + median_mapping_configs
        # inherit
        super()._init_config(tr_data)
        # common
        self._joint_training = self.config.setdefault("joint_training", True)
        self._use_gradient_loss = self.config.setdefault("use_gradient_loss", True)
        fetches = self.config.setdefault("fetches", "all")
        if fetches == "all":
            self._fetches = {"cdf", "quantile"}
        elif isinstance(fetches, str):
            self._fetches = {fetches}
        else:
            self._fetches = set(fetches)
        cdf_ratio_anchors = quantile_anchors = self.default_anchors
        if not self.fetch_cdf:
            cdf_ratio_anchors = None
        if not self.fetch_quantile:
            quantile_anchors = None
        cdf_ratio_anchors = self.config.setdefault("cdf_ratio_anchors", cdf_ratio_anchors)
        quantile_anchors = self.config.setdefault("quantile_anchors", quantile_anchors)
        if cdf_ratio_anchors is None:
            self._cdf_ratio_anchors = None
        else:
            self._cdf_ratio_anchors = np.asarray(cdf_ratio_anchors, np.float32)
        if quantile_anchors is None:
            self._quantile_anchors = None
        else:
            self._quantile_anchors = np.asarray(quantile_anchors, np.float32)
        # use negative value to disable dual inference
        # use 0 (sample) for stable performance & 1 (gradient descent) for faster (and maybe better) performance
        self._dual_inference_version = self.config.setdefault("dual_inference_version", -1)
        self._synthetic_range = self.config.setdefault("synthetic_range", 3)
        # loss config
        self._loss_config = self.config.setdefault("loss_config", {})
        self._loss_config.setdefault("use_dynamic_weights", False)
        self._loss_config.setdefault("use_anneal", self.num_feature_layers > 0)
        self._loss_config.setdefault("mtl_method", None)
        self._loss_config["joint_training"] = self._joint_training
        # pipeline config
        default_metric_types = ["ddr", "loss"] if self.fetch_quantile else ["mae", "loss"]
        pipeline_config = self._wrapper_config.setdefault("pipeline_config", {})
        pipeline_config = update_dict(pipeline_config, {
            "clip_norm": 1., "ema_decay": 0.,
            "num_epoch": 40, "max_epoch": 1000, "batch_size": 128,
            "metric_config": {"types": default_metric_types, "decay": 0.5}
        })
        num_train_samples = tr_data.processed.x.shape[0]
        num_epoch = pipeline_config["num_epoch"]
        batch_size = pipeline_config["batch_size"]
        anneal_step = self._loss_config.setdefault(
            "anneal_step", (num_train_samples * num_epoch) // (batch_size * 2))
        self._loss_config.setdefault("anneal_step", anneal_step)
        self._pipeline_config = self._wrapper_config["pipeline_config"] = pipeline_config
        # optimize schema
        self._reg_step = int(self.config.setdefault("reg_step", 10))
        self._feature_step = int(self.config.setdefault("feature_step", 5))
        self._synthetic_step = int(self.config.setdefault("synthetic_step", 5))

    def _init_loss(self,
                   tr_data: TabularData):
        self.loss = DDRLoss(self._loss_config, self.device)

    def _init_mlp_config(self,
                         prefix: str,
                         default_bias: bool,
                         default_activation: str,
                         units: List[int]) -> Tuple[List[Dict[str, Any]], Dict[str, Any]]:
        activation = self.config.setdefault(f"{prefix}_activation", default_activation)
        mapping_configs_key = f"{prefix}_mapping_configs"
        mapping_configs = self.config.setdefault(
            mapping_configs_key, shallow_copy_dict(self._common_configs))
        if isinstance(mapping_configs, dict):
            mapping_configs.setdefault("bias", default_bias)
            mapping_configs.setdefault("activation", activation)
            mapping_configs = self.config[mapping_configs_key] = [mapping_configs] * len(units)
        final_mapping_config = self.config.setdefault(f"{prefix}_final_mapping_config", {})
        return mapping_configs, final_mapping_config

    def _inject_median_params(self):
        for mapping in self.mlp.mappings[:self.num_feature_layers]:
            self.__feature_params.extend(mapping.parameters())
        for mapping in self.mlp.mappings[self.num_feature_layers:]:
            self.__reg_params.extend(mapping.parameters())

    def _init_cdf_layers(self):
        if not self.fetch_cdf:
            self.cdf_reg = self.cdf_feature_projection = None
            return
        cdf_input_dim = self.merged_dim + 1
        if self.num_feature_layers > 0:
            cdf_input_dim += self.feature_dim
        # regression part
        cdf_reg_units = self.config.setdefault("cdf_reg_units", [512])
        if not cdf_reg_units:
            self.cdf_reg = Linear(cdf_input_dim, 1, **self._common_configs)
        else:
            mapping_configs, final_mapping_config = self._init_mlp_config(
                "cdf_reg", True, "ReLU", cdf_reg_units)
            self.cdf_reg = MLP(
                cdf_input_dim, 1, cdf_reg_units, mapping_configs,
                final_mapping_config=final_mapping_config
            )
        self.__reg_params.extend(self.cdf_reg.parameters())
        # feature part
        if self.no_joint_features:
            self.cdf_feature_projection = None
        else:
            cdf_feature_units = self.feature_units.copy()
            mapping_configs, final_mapping_config = self._init_mlp_config(
                "cdf_feature", False, "ReLU", cdf_feature_units)
            self.cdf_feature_projection = MLP(
                1, None, cdf_feature_units, mapping_configs,
                final_mapping_config=final_mapping_config
            )
            self.__feature_params.extend(self.cdf_feature_projection.parameters())

    def _init_quantile_layers(self):
        if not self.fetch_quantile:
            self.median_residual_reg = None
            self.additive_q_res_reg = self.multiply_q_res_reg = None
            return
        quantile_input_dim = self.merged_dim
        if self.num_feature_layers > 0:
            quantile_input_dim += self.feature_dim
        activations = Activations()
        self.mish = activations.mish
        self.relu = activations.ReLU
        # median residual part
        median_residual_units = self.config.setdefault("median_residual_units", [512])
        mapping_configs, final_mapping_config = self._init_mlp_config(
            "median_residual", True, self._q_reg_activation, median_residual_units)
        self.median_residual_reg = MLP(
            quantile_input_dim, 2, median_residual_units, mapping_configs,
            final_mapping_config=final_mapping_config
        )
        self.__reg_params.extend(self.median_residual_reg.parameters())
        # quantile residual regression part
        qr_input_dim = quantile_input_dim + 1
        qr_reg_units = self.config.setdefault("quantile_res_reg_units", [512])
        mapping_configs, final_mapping_config = self._init_mlp_config(
            "quantile_res_reg", True, self._q_reg_activation, qr_reg_units)
        q_res_reg_args = qr_input_dim, 2, qr_reg_units, mapping_configs, final_mapping_config
        self.additive_q_res_reg = self._make_projection(*q_res_reg_args)
        self.multiply_q_res_reg = self._make_projection(*q_res_reg_args)
        self.__reg_params.extend(self.additive_q_res_reg.parameters())
        self.__reg_params.extend(self.multiply_q_res_reg.parameters())
        # feature part
        if self.no_joint_features:
            self.additive_q_projection = self.multiply_q_projection = None
        else:
            quantile_feature_units = self.feature_units.copy()
            mapping_configs, final_mapping_config = self._init_mlp_config(
                "quantile_feature", False, "mish", quantile_feature_units)
            q_feature_proj_args = 1, None, quantile_feature_units, mapping_configs, final_mapping_config
            self.additive_q_projection = self._make_projection(*q_feature_proj_args)
            self.multiply_q_projection = self._make_projection(*q_feature_proj_args)
            self.__feature_params.extend(self.additive_q_projection.parameters())
            self.__feature_params.extend(self.multiply_q_projection.parameters())

    def _init_parameters(self):
        all_parameters = set(self.parameters())
        self.__base_params = list(all_parameters - set(self.__feature_params) - set(self.__reg_params))
        base_config = {"optimizer": "adam", "optimizer_config": {"lr": 3.5e-4}}
        optimizers = {"all": base_config}
        if self._feature_step > 0:
            optimizers["feature_parameters"] = shallow_copy_dict(base_config)
        if self._reg_step > 0:
            optimizers["reg_parameters"] = shallow_copy_dict(base_config)
        if self.__base_params:
            optimizers["base_params"] = shallow_copy_dict(base_config)
        pipeline_optimizers = self._pipeline_config.setdefault("optimizers", {})
        pipeline_optimizers = update_dict(pipeline_optimizers, optimizers)
        self._pipeline_config["optimizers"] = pipeline_optimizers

    def _optimizer_step(self,
                        optimizers):
        for key in self.target_parameters:
            opt = optimizers[key]
            opt.step()
            opt.zero_grad()

    @staticmethod
    def _make_projection(in_dim: int,
                         out_dim: int,
                         units: List[int],
                         mapping_configs: List[Dict[str, Any]],
                         final_mapping_config: Dict[str, Any]) -> Union[Linear, MLP]:
        if not units:
            return Linear(in_dim, out_dim, **final_mapping_config)
        return MLP(
            in_dim, out_dim, units, mapping_configs,
            final_mapping_config=final_mapping_config
        )

    @property
    def feature_dim(self) -> int:
        return self.feature_units[-1]

    @property
    def num_feature_layers(self) -> int:
        return len(self.feature_units)

    @property
    def no_joint_features(self) -> bool:
        return self.num_feature_layers == 0

    @property
    def quantile_residual_core_keys(self) -> List[str]:
        return ["median_residual", "quantile_residual", "quantile_sign"]

    @property
    def default_anchors(self) -> np.ndarray:
        return np.linspace(0.05, 0.95, 10).astype(np.float32)

    @property
    def fetch_cdf(self) -> bool:
        return "cdf" in self._fetches

    @property
    def fetch_quantile(self) -> bool:
        return "quantile" in self._fetches

    @property
    def trigger_feature_update(self) -> bool:
        return self._feature_step > 0 and self._step_count % self._feature_step == 0

    @property
    def trigger_reg_update(self) -> bool:
        return self._reg_step > 0 and self._step_count % self._reg_step == 0

    @property
    def base_parameters(self) -> List[nn.Parameter]:
        return self.__base_params

    @property
    def feature_parameters(self) -> List[nn.Parameter]:
        return self.__feature_params

    @property
    def reg_parameters(self) -> List[nn.Parameter]:
        return self.__reg_params

    @property
    def target_parameters(self) -> Set[str]:
        if self.trigger_feature_update or self.trigger_reg_update:
            if not self.trigger_feature_update:
                return {"reg_parameters"}
            if not self.trigger_reg_update:
                return {"feature_parameters"}
            return {"feature_parameters", "reg_parameters"}
        return {"all"}

    # build outputs

    @staticmethod
    def _merge_responses(anchors: torch.Tensor,
                         projection: Union[MLP, None],
                         feature_layers: List[torch.Tensor]) -> torch.Tensor:
        net = anchors
        if projection is not None:
            for i, mapping in enumerate(projection.mappings):
                if i > 0:
                    net = net + feature_layers[i]
                net = mapping(net)
        return net

    def _build_cdf(self,
                   feature_layers: List[torch.Tensor],
                   anchor_batch: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        anchor_batch_ratio = self._get_anchor_batch_ratio(anchor_batch)
        net = self._merge_responses(anchor_batch_ratio, self.cdf_feature_projection, feature_layers)
        concat_list = [net, anchor_batch_ratio]
        if self.num_feature_layers > 0:
            concat_list.append(feature_layers[0])
        features = torch.cat(concat_list, dim=1)
        cdf_raw = self.cdf_reg(features)
        cdf = torch.sigmoid(cdf_raw)
        return cdf, cdf_raw

    def _build_median_residual(self,
                               feature_layers: List[torch.Tensor],
                               quantile_batch: Union[torch.Tensor, None],
                               sign: torch.Tensor = None) -> tensor_dict_type:
        concat_list = [feature_layers[-1]]
        if self.num_feature_layers > 0:
            concat_list.append(feature_layers[0])
        mr_features = torch.cat(concat_list, dim=1)
        mr = self.median_residual_reg(mr_features)
        mr = self.mish(mr)
        mr_pos, mr_neg = torch.chunk(mr, 2, dim=1)
        mr_neg = -mr_neg
        if sign is None:
            sign = torch.sign(quantile_batch - 0.5)
        pos_quantile_mask = sign == 1.
        median_residual = torch.where(pos_quantile_mask, mr_pos, mr_neg)
        return {
            "quantile_sign": sign,
            "median_residual": median_residual,
            "pos_quantile_mask": pos_quantile_mask
        }

    def _build_quantile_residual(self,
                                 feature_layers: List[torch.Tensor],
                                 quantile_batch: torch.Tensor,
                                 pressure_batch: bool = False,
                                 *,
                                 keys: Union[str, List[str]] = None) -> tensors_type:
        add_ratio, mul_ratio = self._get_quantile_batch_ratio(quantile_batch)
        add_net = self._merge_responses(add_ratio, self.additive_q_projection, feature_layers)
        mul_net = self._merge_responses(mul_ratio, self.multiply_q_projection, feature_layers)
        concat_list = []
        if self.num_feature_layers > 0:
            concat_list.append(feature_layers[0])
        sub_quantile_dict, sub_quantile_pos_dict, sub_quantile_neg_dict = {}, {}, {}
        mr_results = self._build_median_residual(feature_layers, quantile_batch)
        median_residual, q_sign, q_pos_mask = map(
            mr_results.get, ["median_residual", "quantile_sign", "pos_quantile_mask"])
        for res_type, q_features, q_ratio, q_res_reg, activation in zip(
            ["add", "mul"], [add_net, mul_net], [add_ratio, mul_ratio],
            [self.additive_q_res_reg, self.multiply_q_res_reg], [self.mish, self.relu]
        ):
            local_concat_list = concat_list + [q_features, q_ratio]
            local_quantile_features = torch.cat(local_concat_list, dim=1)
            sq = q_res_reg(local_quantile_features)
            sq = activation(sq)
            sub_quantile_pos, sub_quantile_neg = torch.chunk(sq, 2, dim=1)
            if res_type == "add":
                sub_quantile_neg = -sub_quantile_neg
            if pressure_batch:
                sub_quantile_pos_dict[res_type] = sub_quantile_pos
                sub_quantile_neg_dict[res_type] = sub_quantile_neg
                continue
            sub_quantile_dict[res_type] = torch.where(q_pos_mask, sub_quantile_pos, sub_quantile_neg)
        if pressure_batch:
            return {
                "pressure_sub_quantile_pos_dict": sub_quantile_pos_dict,
                "pressure_sub_quantile_neg_dict": sub_quantile_neg_dict
            }
        quantile_residual = median_residual.detach() * sub_quantile_dict["mul"] + sub_quantile_dict["add"]
        rs = {
            "quantile_sign": q_sign,
            "median_residual": median_residual, "quantile_residual": quantile_residual,
            "additive_quantile_features": add_net, "multiply_quantile_features": mul_net,
        }
        if keys is None:
            return rs
        if isinstance(keys, str):
            return rs[keys]
        return [rs[key] for key in keys]

    # fetch & combine outputs

    def _get_anchor_batch_ratio(self,
                                anchor_batch: torch.Tensor) -> torch.Tensor:
        return (anchor_batch - self.y_min) / self.y_diff

    @staticmethod
    def _get_quantile_batch_ratio(quantile_batch: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        add_ratio = 4 * torch.where(quantile_batch > 0.5, quantile_batch - 0.75, 0.25 - quantile_batch)
        mul_ratio = 2 * torch.abs(quantile_batch - 0.5)
        return add_ratio, mul_ratio

    def _convert_np_anchors(self,
                            np_anchors: Union[np.ndarray, None]) -> Union[torch.Tensor, None]:
        if np_anchors is None:
            return
        return to_torch(np_anchors.reshape([-1, 1])).to(self.device).requires_grad_(True)

    def _get_feature_layers(self,
                            net: torch.Tensor) -> List[torch.Tensor]:
        feature_layers = [net]
        for mapping in self.mlp.mappings[:self.num_feature_layers]:
            net = mapping(net)
            feature_layers.append(net)
        return feature_layers

    def _get_median(self,
                    feature_layers: List[torch.Tensor]) -> torch.Tensor:
        net = feature_layers[-1]
        for mapping in self.mlp.mappings[self.num_feature_layers:]:
            net = mapping(net)
        return net

    def _get_cdf(self,
                 basic_info: BasicInfo,
                 fetch_anchors: bool) -> tensor_dict_type:
        if not self.fetch_cdf:
            raise ValueError("cdf could not be fetched")
        anchor_batch = basic_info.anchor_batch
        feature_layers = basic_info.feature_layers
        cdf, cdf_raw = self._build_cdf(feature_layers, anchor_batch)
        # get specific sampled cdf
        if not fetch_anchors or self._cdf_ratio_anchors is None or not self.training:
            sampled_anchors = sampled_cdf = sampled_cdf_raw = None
        else:
            choice_indices = np.random.randint(0, len(self._cdf_ratio_anchors), len(cdf))
            choice_array = self.y_min + (self.y_max - self.y_min) * self._cdf_ratio_anchors
            sampled_anchors = self._convert_np_anchors(choice_array[choice_indices])
            sampled_cdf, sampled_cdf_raw = self._build_cdf(feature_layers, sampled_anchors)
        return {
            "cdf": cdf, "cdf_raw": cdf_raw, "sampled_anchors": sampled_anchors,
            "sampled_cdf": sampled_cdf, "sampled_cdf_raw": sampled_cdf_raw
        }

    def _get_quantile_residual(self,
                               basic_info: BasicInfo,
                               fetch_anchors: bool) -> tensor_dict_type:
        if not self.fetch_quantile:
            raise ValueError("quantile could not be fetched")
        quantile_batch = basic_info.quantile_batch
        feature_layers = basic_info.feature_layers
        quantile_dict = self._build_quantile_residual(feature_layers, quantile_batch)
        # get median pressure
        with torch.no_grad():
            median_pressure_batch = quantile_batch.new_empty(quantile_batch.shape).fill_(0.5)
        median_pressure_batch.requires_grad_(True)
        quantile_dict.update(self._build_quantile_residual(
            feature_layers, median_pressure_batch, pressure_batch=True))
        # get specific sampled quantile
        if not fetch_anchors or self._quantile_anchors is None or not self.training:
            sampled_quantiles = smr = sqr = sqs = None
        else:
            choice_indices = np.random.randint(0, len(self._quantile_anchors), len(quantile_batch))
            sampled_quantiles = self._convert_np_anchors(self._quantile_anchors[choice_indices])
            smr, sqr, sqs = self._build_quantile_residual(
                feature_layers, sampled_quantiles, keys=self.quantile_residual_core_keys)
        quantile_dict.update({
            "sampled_quantiles": sampled_quantiles, "sampled_median_residual": smr,
            "sampled_quantile_residual": sqr, "sampled_quantile_sign": sqs
        })
        return quantile_dict

    def _get_dual_cdf(self,
                      basic_info: BasicInfo,
                      quantile_residual: torch.Tensor) -> tensor_dict_type:
        median_detach = basic_info.median_detach
        feature_layers = basic_info.feature_layers
        dual_cdf, dual_cdf_raw = self._build_cdf(feature_layers, quantile_residual + median_detach)
        cmr, cdf_quantile_residual, cqs = self._build_quantile_residual(
            feature_layers, dual_cdf, keys=self.quantile_residual_core_keys)
        return {
            "dual_cdf": dual_cdf, "dual_cdf_raw": dual_cdf_raw, "cdf_median_residual": cmr,
            "cdf_quantile_residual": cdf_quantile_residual, "cdf_quantile_sign": cqs
        }

    def _get_dual_quantile(self,
                           basic_info: BasicInfo,
                           cdf: torch.Tensor) -> tensor_dict_type:
        median_detach = basic_info.median_detach
        feature_layers = basic_info.feature_layers
        dmr, dual_quantile_residual, dqs = self._build_quantile_residual(
            feature_layers, cdf, keys=self.quantile_residual_core_keys)
        dual_quantile = median_detach + dual_quantile_residual
        quantile_cdf, quantile_cdf_raw = self._build_cdf(feature_layers, dual_quantile)
        return {
            "dual_median_residual": dmr, "dual_quantile": dual_quantile, "dual_quantile_residual": dqs,
            "quantile_cdf": quantile_cdf, "quantile_cdf_raw": quantile_cdf_raw
        }

    # predict methods

    def _predict_cdf(self,
                     init: torch.Tensor,
                     y_batch: torch.Tensor) -> torch.Tensor:
        feature_layers = self._get_feature_layers(init)
        cdf = self._build_cdf(feature_layers, y_batch)[0]
        return cdf

    def _predict_pdf(self,
                     init: torch.Tensor,
                     y_batch: torch.Tensor) -> torch.Tensor:
        feature_layers = self._get_feature_layers(init)
        cdf = self._build_cdf(feature_layers, y_batch.requires_grad_(True))[0]
        pdf = get_gradient(cdf, y_batch)
        y_batch.requires_grad_(False)
        return pdf

    def _predict_quantile(self,
                          init: torch.Tensor,
                          q_batch: torch.Tensor) -> torch.Tensor:
        feature_layers = self._get_feature_layers(init)
        median = self._get_median(feature_layers)
        quantile = median + self._build_quantile_residual(
            feature_layers, q_batch, keys="quantile_residual")
        return quantile

    def _predict_median_residual(self,
                                 init: torch.Tensor,
                                 sign_batch: torch.Tensor) -> torch.Tensor:
        feature_layers = self._get_feature_layers(init)
        median = self._get_median(feature_layers)
        mr = self._build_median_residual(feature_layers, None, sign_batch)["median_residual"]
        residual = median + mr
        return residual

    # Core

    def _switch_training_status(self) -> None:
        target = self.target_parameters
        if "all" in target:
            self._switch_requires_grad(self.__base_params, True)
            self._switch_requires_grad(self.__feature_params, True)
            self._switch_requires_grad(self.__reg_params, True)
        else:
            self._switch_requires_grad(self.__base_params, False)
            self._switch_requires_grad(self.__feature_params, "feature_parameters" in target)
            self._switch_requires_grad(self.__reg_params, "reg_parameters" in target)

    def _expand(self,
                n: int,
                elem: Union[float, torch.Tensor],
                *,
                numpy: bool = False,
                to_device: bool = True) -> Union[torch.Tensor, np.ndarray]:
        if isinstance(elem, torch.Tensor):
            return elem
        if not is_numeric(elem):
            elem = np.asarray(elem, np.float32)
        else:
            elem = np.repeat(elem, n).astype(np.float32)
        elem = elem.reshape([-1, 1])
        if numpy:
            return elem
        elem = torch.from_numpy(elem)
        if to_device:
            elem = elem.to(self.device)
        return elem

    def _core(self,
              init: torch.Tensor,
              **kwargs) -> tensor_dict_type:
        # median only
        if kwargs.get("median_only", False):
            with timing_context(self, "forward.no_loss"):
                with eval_context(self):
                    feature_layers = self._get_feature_layers(init)
                    median = self._get_median(feature_layers)
                    return {"predictions": median}
        # generate quantile / anchor batch
        with timing_context(self, "forward.generate_anchors"):
            if self.training:
                if not self.fetch_quantile:
                    quantile_batch = None
                else:
                    quantile_batch = np.random.random([len(init), 1]).astype(np.float32)
                if not self.fetch_cdf:
                    anchor_batch = None
                else:
                    anchor_batch = np.random.random([len(init), 1]).astype(np.float32)
                    anchor_batch *= (self.y_max - self.y_min) + self.y_min
            else:
                if not self.fetch_cdf and not self.fetch_quantile:
                    quantile_batch = None
                else:
                    quantile_batch = self.default_anchors
                    n_repeat = int(len(init) / len(quantile_batch)) + 1
                    quantile_batch = np.repeat(quantile_batch, n_repeat)[:len(init)]
                if not self.fetch_cdf:
                    anchor_batch = None
                else:
                    anchor_batch = quantile_batch * (self.y_max - self.y_min) + self.y_min
            anchor_batch = self._convert_np_anchors(anchor_batch)
            quantile_batch = self._convert_np_anchors(quantile_batch)
        # build features
        with timing_context(self, "forward.build_features"):
            feature_layers = self._get_feature_layers(init)
        # build median
        with timing_context(self, "forward.median"):
            median = self._get_median(feature_layers)
        # construct basic info
        basic_info = BasicInfo(
            median, median.detach(),
            feature_layers, anchor_batch, quantile_batch
        )
        # is synthetic
        is_synthetic = kwargs.get("synthetic", False)
        # build cdf
        if not self.fetch_cdf:
            cdf_dict = {}
            cdf_gradient = sampled_cdf_gradient = None
        else:
            with timing_context(self, "forward.cdf"):
                cdf_dict = self._get_cdf(basic_info, self.training and not is_synthetic)
            # build cdf gradients
            sampled_cdf_gradient = None
            cdf_raw, sampled_anchors, sampled_cdf_raw = map(
                cdf_dict.get, ["cdf_raw", "sampled_anchors", "sampled_cdf_raw"])
            fetch_cdf_gradient = self._use_gradient_loss and self.training
            if not is_synthetic and not fetch_cdf_gradient:
                cdf_gradient = None
            else:
                with timing_context(self, "forward.cdf_gradient"):
                    cdf_gradient = get_gradient(cdf_raw, anchor_batch, True, True)
                    if sampled_cdf_raw is not None:
                        sampled_cdf_gradient = get_gradient(sampled_cdf_raw, sampled_anchors, True, True)
        # build quantile
        if not self.fetch_quantile:
            quantile_dict = {}
            quantile_residual = None
            quantile_residual_gradient = sampled_qr_gradient = None
        else:
            with timing_context(self, "forward.quantile"):
                quantile_dict = self._get_quantile_residual(basic_info, self.training and not is_synthetic)
            # build quantile gradients
            sampled_qr_gradient = None
            quantile_residual, sampled_quantiles, sqr = map(
                quantile_dict.get, ["quantile_residual", "sampled_quantiles", "sampled_quantile_residual"])
            fetch_quantile_gradient = self._use_gradient_loss and self.training
            if not is_synthetic and not fetch_quantile_gradient:
                quantile_residual_gradient = None
            else:
                with timing_context(self, "forward.quantile_gradient"):
                    quantile_residual_gradient = get_gradient(quantile_residual, quantile_batch, True, True)
                    if sqr is not None:
                        sampled_qr_gradient = get_gradient(sqr, sampled_quantiles, True, True)
        # build dual
        dual_ca_gradient = dual_cq_gradient = None
        dual_qa_gradient = dual_qq_gradient = None
        if not self._joint_training or not self.fetch_cdf or not self.fetch_quantile or is_synthetic:
            dual_cdf_dict, dual_quantile_dict = {}, {}
        else:
            with timing_context(self, "forward.dual_cdf"):
                dual_cdf_dict = self._get_dual_cdf(basic_info, quantile_residual)
            with timing_context(self, "forward.dual_quantile"):
                dual_quantile_dict = self._get_dual_quantile(basic_info, cdf_dict["cdf"])
        # construct results
        rs = {
            "init": init, "predictions": median,
            "median_detach": basic_info.median_detach, "feature_layers": feature_layers,
            "anchor_batch": anchor_batch, "quantile_batch": quantile_batch,
            "cdf_gradient": cdf_gradient, "quantile_residual_gradient": quantile_residual_gradient,
            "sampled_cdf_gradient": sampled_cdf_gradient, "sampled_qr_gradient": sampled_qr_gradient,
            "dual_ca_gradient": dual_ca_gradient, "dual_cq_gradient": dual_cq_gradient,
            "dual_qa_gradient": dual_qa_gradient, "dual_qq_gradient": dual_qq_gradient
        }
        for d in (cdf_dict, quantile_dict, dual_cdf_dict, dual_quantile_dict):
            rs.update(d)
        return rs

    # API

    def forward(self,
                batch: tensor_dict_type,
                **kwargs) -> tensor_dict_type:
        forward_dict = {}
        x_batch = batch["x_batch"]
        init = self._split_features(x_batch).merge()
        predict_pdf, predict_cdf = map(kwargs.get, ["predict_pdf", "predict_cdf"])
        predict_quantile, predict_median_residual = map(
            kwargs.get, ["predict_quantile", "predict_median_residual"])
        if predict_pdf or predict_cdf:
            y = kwargs.get("y")
            if y is None:
                raise ValueError(f"pdf / cdf cannot be predicted without y")
            y_batch = self._expand(len(init), y, numpy=True)
            y_batch = self.tr_data.transform_labels(y_batch)
            y_batch = to_torch(y_batch).to(self.device)
            if predict_pdf:
                forward_dict["pdf"] = self._predict_pdf(init, y_batch)
            if predict_cdf:
                forward_dict["cdf"] = self._predict_cdf(init, y_batch)
        if predict_quantile:
            q = kwargs.get("q")
            if q is None:
                raise ValueError(f"quantile cannot be predicted without q")
            q_batch = self._expand(len(init), q)
            forward_dict["quantile"] = self._predict_quantile(init, q_batch)
        if predict_median_residual:
            expand = lambda sign: self._expand(len(init), sign)
            sign_batch = torch.cat(list(map(expand, [1, -1])))
            init2 = init.repeat([2, 1])
            pos, neg = torch.chunk(self._predict_median_residual(init2, sign_batch), 2, dim=0)
            forward_dict["mr_pos"] = pos
            forward_dict["mr_neg"] = neg
        if not forward_dict:
            forward_dict = self._core(init, **kwargs)
        return forward_dict

    def loss_function(self,
                      batch: tensor_dict_type,
                      forward_results: tensor_dict_type) -> Dict[str, Union[torch.Tensor, float]]:
        y_batch = batch["y_batch"]
        init = forward_results["init"]
        loss, loss_dict = self.loss(forward_results, y_batch)
        if self.training and self._synthetic_step > 0 and self._step_count % self._synthetic_step == 0:
            with timing_context(self, "synthetic.get_batch"):
                x_min, x_max = torch.min(init, dim=0)[0], torch.max(init, dim=0)[0]
                x_diff = x_max - x_min
                lower_bound = 0.5 * (self._synthetic_range - 1) * x_diff
                synthetic_x_batch = init.new_empty(init.shape).uniform_(
                    0, 1) * x_diff * self._synthetic_range - (lower_bound - x_min)
            with timing_context(self, "synthetic.forward"):
                synthetic_outputs = self.forward(
                    {"x_batch": synthetic_x_batch},
                    no_loss=False, synthetic=True
                )["predictions"]
            with timing_context(self, "synthetic.loss"):
                synthetic_loss, synthetic_loss_info = self.loss(
                    {"predictions": synthetic_outputs}, y_batch,
                    check_monotonous_only=True
                )
            loss_dict["synthetic"] = synthetic_loss
            loss = loss + synthetic_loss
        loss_dict["loss"] = loss
        if self.training:
            self._step_count += 1
        else:
            quantile_losses = []
            y_batch = to_numpy(y_batch)
            for q in self._quantile_anchors:
                self.q_metric.config["q"] = q
                yq = self._predict_quantile(init, self._expand(len(init), q))
                quantile_losses.append(self.q_metric.metric(y_batch, to_numpy(yq)))
            quantile_metric = -sum(quantile_losses) / len(quantile_losses) * self.q_metric.sign
            loss_dict["ddr"] = torch.tensor([quantile_metric], dtype=torch.float32)
        return loss_dict


__all__ = ["DDR"]



import torch.optim.lr_scheduler as scheduler

scheduler_dict = {}


def register_scheduler(name):
    def _register(cls_):
        global scheduler_dict
        scheduler_dict[name] = cls_
        return cls_
    return _register


register_scheduler("plateau")(scheduler.ReduceLROnPlateau)


@register_scheduler("warmup")
class WarmupScheduler(scheduler._LRScheduler):
    def __init__(self, optimizer, multiplier, warmup_step,
                 scheduler_afterwards_base=None, scheduler_afterwards_config=None):
        self.multiplier = multiplier
        assert self.multiplier > 1., "multiplier should be greater than 1"
        self.warmup_step, self.finished_warmup = warmup_step, False
        if scheduler_afterwards_config is None:
            scheduler_afterwards_config = {}
        self.scheduler_afterwards = scheduler_afterwards_base(optimizer, **scheduler_afterwards_config)
        super().__init__(optimizer)

    @property
    def lr_warmup_func(self):
        return lambda lr: lr * ((self.multiplier - 1.) * self.last_epoch / self.warmup_step + 1.)

    @property
    def lr_multiplier_func(self):
        return lambda lr: lr * self.multiplier

    @property
    def reduce_on_plateau_afterwards(self):
        return isinstance(self, scheduler.ReduceLROnPlateau)

    def _step_reduce_on_plateau(self, metrics, epoch=None):
        if epoch is None:
            epoch = self.last_epoch + 1
        self.last_epoch = epoch
        if self.last_epoch <= self.warmup_step:
            for param_group, lr in zip(self.optimizer.param_groups, map(self.lr_warmup_func, self.base_lrs)):
                param_group["lr"] = lr
        else:
            if epoch is not None:
                epoch -= self.warmup_step
            self.scheduler_afterwards.step(metrics, None)

    def get_lr(self):
        if self.last_epoch > self.warmup_step:
            if self.scheduler_afterwards is not None:
                if not self.finished_warmup:
                    self.finished_warmup = True
                    self.scheduler_afterwards.base_lrs = list(map(self.lr_multiplier_func, self.base_lrs))
                return self.scheduler_afterwards.get_lr()
            return list(map(self.lr_multiplier_func, self.base_lrs))
        return list(map(self.lr_warmup_func, self.base_lrs))

    def step(self, epoch=None, metrics=None):
        if self.reduce_on_plateau_afterwards:
            self._step_reduce_on_plateau(metrics, epoch)
        else:
            if not self.finished_warmup or self.scheduler_afterwards is None:
                return super().step(epoch)
            if epoch is not None:
                epoch -= self.warmup_step
            self.scheduler_afterwards.step(epoch)


__all__ = ["scheduler_dict", "register_scheduler"]



import math
import torch
import logging

import torch.nn as nn

from typing import *
from cftool.misc import *
from abc import ABCMeta, abstractmethod

from ..misc.toolkit import tensor_dict_type

encoder_dict: Dict[str, Type["EncoderBase"]] = {}


class EncoderBase(nn.Module, LoggingMixin, metaclass=ABCMeta):
    def __init__(self,
                 idx: int,
                 num_values: int,
                 config: Dict[str, Any]):
        super().__init__()
        self.idx, self.num_values = idx, num_values
        self._init_config(config)

    def _init_config(self, config: Dict[str, Any]):
        self.config = config

    @property
    @abstractmethod
    def dim(self) -> int:
        pass

    @abstractmethod
    def _core(self,
              selected: torch.Tensor) -> torch.Tensor:
        pass

    def forward(self,
                categorical_column: torch.Tensor) -> torch.Tensor:
        selected = categorical_column.to(torch.long)
        # TODO : cache oob_masks for static datasets
        oob_mask = selected >= self.num_values
        if torch.any(oob_mask):
            self.log_msg(
                f"out of bound occurred in categorical column {self.idx}, "
                f"ratio : {torch.mean(oob_mask.to(torch.float)).item():8.6f}",
                self.warning_prefix, 5, logging.WARNING
            )
            selected[oob_mask] = 0
        return self._core(selected)

    @classmethod
    def register(cls, name: str):
        global encoder_dict
        def before(cls_): cls_.__identifier__ = name
        return register_core(name, encoder_dict, before_register=before)


@EncoderBase.register("one_hot")
class OneHot(EncoderBase):
    @property
    def dim(self) -> int:
        return self.num_values

    def _core(self,
              selected: torch.Tensor) -> torch.Tensor:
        return nn.functional.one_hot(selected, num_classes=self.num_values).to(torch.float32)


@EncoderBase.register("embedding")
class Embedding(EncoderBase):
    def __init__(self,
                 idx: int,
                 num_values: int,
                 config: Dict[str, Any]):
        super().__init__(idx, num_values, config)
        self.embedding = nn.Embedding(num_values, self._dim)

    def _init_config(self, config: Dict[str, Any]):
        super()._init_config(config)
        self._mean = self.config.setdefault("embedding_mean", 0.)
        self._std = self.config.setdefault("embedding_std", 0.02)
        embedding_dim = self.config.setdefault("embedding_dim", "auto")
        if isinstance(embedding_dim, int):
            self._dim = embedding_dim
        elif embedding_dim == "log":
            self._dim = math.ceil(math.log2(self.num_values))
        elif embedding_dim == "sqrt":
            self._dim = math.ceil(math.sqrt(self.num_values))
        elif embedding_dim == "auto":
            self._dim = min(self.num_values, max(4, min(8, math.ceil(math.log2(self.num_values)))))
        else:
            raise ValueError(f"embedding dim '{embedding_dim}' is not defined")

    @property
    def dim(self) -> int:
        return self._dim

    def _core(self,
              selected: torch.Tensor) -> torch.Tensor:
        return self.embedding(selected)


class EncoderStack(nn.Module, LoggingMixin):
    def __init__(self, *encoders):
        super().__init__()
        encoders_ = {}
        for encoder in encoders:
            key = encoder.__identifier__
            if key in encoders_:
                raise ValueError(f"'{key}' encoder is already stacked")
            encoders_[key] = encoder
        self.encoders = nn.ModuleDict(encoders_)
        self.sorted_keys = sorted(encoders_)

    @property
    def dim(self) -> int:
        return sum(self.dims.values())

    @property
    def dims(self) -> Dict[str, int]:
        return {k: encoder.dim for k, encoder in self.encoders.items()}

    def forward(self,
                categorical_column: torch.Tensor,
                *,
                return_all: bool = False) -> Union[torch.Tensor, tensor_dict_type]:
        encodings = {k: v(categorical_column) for k, v in self.encoders.items()}
        if return_all:
            return encodings
        return torch.cat([encodings[k] for k in self.sorted_keys], dim=1)


__all__ = ["EncoderBase", "EncoderStack", "encoder_dict"]



import os
import json
import torch
import pprint
import shutil
import inspect
import logging

import numpy as np
import torch.nn as nn

from typing import *
from cftool.ml import *
from cftool.misc import *
from cfdata.tabular import *
from tqdm import tqdm
from functools import partial

from abc import ABCMeta, abstractmethod

from .modules import *
from .misc.toolkit import *

model_dict: Dict[str, Type["ModelBase"]] = {}


class SplitFeatures(NamedTuple):
    categorical: Union[torch.Tensor, tensor_dict_type, None]
    numerical: Union[torch.Tensor, None]

    @property
    def stacked_categorical(self) -> torch.Tensor:
        if not isinstance(self.categorical, dict):
            return self.categorical
        return torch.cat([self.categorical[k] for k in sorted(self.categorical)], dim=1)

    def merge(self) -> torch.Tensor:
        categorical = self.stacked_categorical
        if categorical is None:
            return self.numerical
        if self.numerical is None:
            return categorical
        return torch.cat([self.numerical, categorical], dim=1)


class ModelBase(nn.Module, LoggingMixin, metaclass=ABCMeta):
    def __init__(self,
                 config: Dict[str, Any],
                 tr_data: TabularData,
                 device: torch.device):
        super().__init__()
        self.device = device
        self._wrapper_config = config
        self.config = config.setdefault("model_config", {})
        self._preset_config(tr_data)
        self._init_config(tr_data)
        self._init_loss(tr_data)
        # encoders
        excluded = 0
        recognizers = tr_data.recognizers
        self.encoders = {}
        self.numerical_columns_mapping = {}
        self.categorical_columns_mapping = {}
        for idx, recognizer in recognizers.items():
            if idx == -1:
                continue
            if not recognizer.info.is_valid:
                excluded += 1
            elif recognizer.info.column_type is ColumnTypes.NUMERICAL:
                self.numerical_columns_mapping[idx] = idx - excluded
            else:
                self._init_encoder(idx)
                self.categorical_columns_mapping[idx] = idx - excluded
        if self.encoders:
            self.encoders = nn.ModuleDict(self.encoders)
        self._categorical_dim = sum(encoder.dim for encoder in self.encoders.values())
        self._numerical_columns = sorted(self.numerical_columns_mapping.values())

    @abstractmethod
    def forward(self,
                batch: tensor_dict_type,
                **kwargs) -> tensor_dict_type:
        # batch will have `categorical`, `numerical` and `labels` keys
        # requires returning `predictions` key
        pass

    def loss_function(self,
                      batch: tensor_dict_type,
                      forward_results: tensor_dict_type) -> Dict[str, Union[torch.Tensor, float]]:
        # requires returning `loss` key
        y_batch = batch["y_batch"]
        if self.tr_data.is_clf:
            y_batch = y_batch.view(-1)
        predictions = forward_results["predictions"]
        return {"loss": self.loss(predictions, y_batch)}

    @property
    def merged_dim(self):
        return self._categorical_dim + len(self._numerical_columns)

    @property
    def encoding_dims(self) -> Dict[str, int]:
        encoding_dims = {}
        for encoder_stack in self.encoders.values():
            dims = encoder_stack.dims
            if not encoding_dims:
                encoding_dims = dims
            else:
                for k, v in dims.items():
                    encoding_dims[k] += v
        return encoding_dims

    @property
    def categorical_dims(self) -> Dict[int, int]:
        dims = {}
        for idx_str in sorted(self.encoders):
            idx = int(idx_str)
            true_idx = self.categorical_columns_mapping[idx]
            dims[true_idx] = self.encoders[idx_str].dim
        return dims

    def _preset_config(self,
                       tr_data: TabularData):
        pass

    def _init_config(self,
                     tr_data: TabularData):
        self.tr_data = tr_data
        # TODO : optimize encodings by pre-calculate one-hot encodings in Wrapper
        self._encoding_methods = self.config.setdefault("encoding_methods", {})
        self._encoding_configs = self.config.setdefault("encoding_configs", {})
        self._default_encoding_method = self.config.setdefault("default_encoding_method", "embedding")

    def _init_loss(self,
                   tr_data: TabularData):
        if tr_data.is_reg:
            self.loss = nn.L1Loss()
        else:
            self.loss = nn.CrossEntropyLoss()

    def _init_encoder(self, idx: int):
        methods = self._encoding_methods.setdefault(idx, self._default_encoding_method)
        config = self._encoding_configs.setdefault(idx, {})
        num_values = self.tr_data.recognizers[idx].num_unique_values
        if isinstance(methods, str):
            methods = [methods]
        encoders = [encoder_dict[method](idx, num_values, config) for method in methods]
        self.encoders[str(idx)] = EncoderStack(*encoders)

    def _optimizer_step(self,
                        optimizers):
        for opt in optimizers.values():
            opt.step()
            opt.zero_grad()

    @staticmethod
    def _switch_requires_grad(params: List[torch.nn.Parameter],
                              requires_grad: bool):
        for param in params:
            param.requires_grad_(requires_grad)

    @staticmethod
    def _collate_tensor_dicts(ds: List[tensor_dict_type],
                              dim: int = 0) -> tensor_dict_type:
        results = {}
        d0 = ds[0]
        for k in d0.keys():
            if not isinstance(d0[k], torch.Tensor):
                continue
            tensors = []
            for rs in ds:
                tensor = rs[k]
                if len(tensor.shape) == 0:
                    tensor = tensor.reshape([1])
                tensors.append(tensor)
            results[k] = torch.cat(tensors, dim=dim)
        return results

    @staticmethod
    def to_prob(raw: np.ndarray) -> np.ndarray:
        return nn.functional.softmax(torch.from_numpy(raw), dim=1).numpy()

    def _split_features(self,
                        x_batch: torch.Tensor,
                        *,
                        return_all_encodings: bool = False) -> SplitFeatures:
        categorical_columns = []
        for idx_str in sorted(self.encoders):
            encoder = self.encoders[idx_str]
            mapping_idx = self.categorical_columns_mapping[int(idx_str)]
            categorical_columns.append(encoder(x_batch[..., mapping_idx], return_all=return_all_encodings))
        if not categorical_columns:
            categorical = None
        elif not return_all_encodings:
            categorical = torch.cat(categorical_columns, dim=1)
        else:
            categorical = self._collate_tensor_dicts(categorical_columns, dim=1)
        numerical = None if not self._numerical_columns else x_batch[..., self._numerical_columns]
        return SplitFeatures(categorical, numerical)

    def get_split(self,
                  processed: np.ndarray,
                  device: torch.device) -> SplitFeatures:
        return self._split_features(torch.from_numpy(processed).to(device))

    @classmethod
    def register(cls, name: str):
        global model_dict
        def before(cls_): cls_.__identifier__ = name
        return register_core(name, model_dict, before_register=before)


class Pipeline(nn.Module, LoggingMixin):
    def __init__(self,
                 model: ModelBase,
                 config: Dict[str, Any],
                 verbose_level: int):
        super().__init__()
        self._init_config(config)
        self.model = model
        self._verbose_level = verbose_level

    def _init_config(self, config):
        self._wrapper_config = config
        self.config = config.setdefault("pipeline_config", {})
        self.batch_size = self.config.setdefault("batch_size", 128)
        self.cv_batch_size = self.config.setdefault("cv_batch_size", 5 * self.batch_size)
        self.use_tqdm = self.config.setdefault("use_tqdm", True)
        self.min_epoch = int(self.config.setdefault("min_epoch", 0))
        self.num_epoch = int(self.config.setdefault("num_epoch", max(40, self.min_epoch)))
        self.max_epoch = int(self.config.setdefault("max_epoch", max(200, self.num_epoch)))
        self.max_snapshot_num = int(self.config.setdefault("max_snapshot_num", 5))
        self.snapshot_start_step = int(self.config.setdefault("snapshot_start_step", 0))
        self._num_step_per_snapshot = int(self.config.setdefault("num_step_per_snapshot", 0))
        self.num_snapshot_per_epoch = int(self.config.setdefault("num_snapshot_per_epoch", 2))
        self.max_step_per_snapshot = int(self.config.setdefault("max_step_per_snapshot", 1000))
        self.plateau_monitor_start_snapshot = int(
            self.config.setdefault("plateau_monitor_start_snapshot", self.num_epoch))

        self._clip_norm = self.config.setdefault("clip_norm", 0.)
        ema_decay = self.config.setdefault("ema_decay", 0.)
        if not 0. < ema_decay < 1.:
            self.ema_decay = None
        else:
            self.ema_decay = EMA(ema_decay, self.model.named_parameters())

        self._logging_path_ = config["_logging_path_"]
        self.logging_folder = config["logging_folder"]
        self.checkpoint_folder = self.config.setdefault(
            "checkpoint_folder",
            os.path.join(self.logging_folder, "checkpoints")
        )
        if os.path.isdir(self.checkpoint_folder):
            self.log_msg(
                f"'{self.checkpoint_folder}' already exists, all of its contents will be removed",
                self.warning_prefix, msg_level=logging.WARNING
            )
            shutil.rmtree(self.checkpoint_folder)

    def _init_data(self, tr_data, cv_data):
        self.tr_data, self.cv_data = tr_data, cv_data
        tr_sampler = ImbalancedSampler(tr_data, verbose_level=self._verbose_level)
        self.tr_loader = DataLoader(self.batch_size, tr_sampler, verbose_level=self._verbose_level)
        if cv_data is None:
            self.cv_loader = None
        else:
            cv_sampler = ImbalancedSampler(cv_data, shuffle=False, verbose_level=self._verbose_level)
            self.cv_loader = DataLoader(self.cv_batch_size, cv_sampler, verbose_level=self._verbose_level)

    def _define_optimizer(self, params_name, optimizer_base, optimizer_config):
        if params_name == "all":
            parameters = self.model.parameters()
        else:
            parameters = getattr(self.model, params_name)
        opt = self.optimizers[params_name] = optimizer_base(parameters, **optimizer_config)
        return opt

    def _init_optimizers(self):
        optimizers = self.config.setdefault("optimizers", {"all": {}})
        self.optimizers, self.schedulers = {}, {}
        for params_name, params_optimizer_config in optimizers.items():
            optimizer = params_optimizer_config.setdefault("optimizer", "adam")
            optimizer_config = params_optimizer_config.setdefault("optimizer_config", {})
            scheduler = params_optimizer_config.setdefault("scheduler", "plateau")
            scheduler_config = params_optimizer_config.setdefault("scheduler_config", {})
            # optimizer
            optimizer_config.setdefault("lr", 1e-3)
            if optimizer == "nag":
                optimizer_config.setdefault("momentum", 0.999)
                optimizer_config.setdefault("weight_decay", 1e-7)
            if scheduler == "warmup":
                multiplier = scheduler_config.setdefault("multiplier", 3)
                default_warm_up_step = min(
                    10 * len(self.tr_loader),
                    int(0.25 * self.plateau_monitor_start_snapshot * self.num_step_per_snapshot)
                )
                warmup_step = scheduler_config.setdefault("warmup_step", default_warm_up_step)
                self.plateau_monitor_start_snapshot += int(warmup_step / self.num_step_per_snapshot)
                self.snapshot_start_step += warmup_step
                optimizer_config["lr"] /= multiplier
            optimizer_base = optimizer_dict[optimizer] if isinstance(optimizer, str) else optimizer
            opt = self._define_optimizer(params_name, optimizer_base, optimizer_config)
            self.config["optimizer_config"] = optimizer_config
            self._optimizer_type = optimizer
            # scheduler
            plateau_default_config: Dict[str, Any] = {"mode": "max"}
            plateau_default_config.setdefault("verbose", self._verbose_level_ >= 3)
            plateau_default_config.setdefault(
                "patience",
                max(10, self.snapshot_start_step // self.num_step_per_snapshot)
            )
            if scheduler == "plateau":
                scheduler_config = update_dict(scheduler_config, plateau_default_config)
            elif scheduler == "warmup":
                scheduler_afterwards_base = scheduler_config.get("scheduler_afterwards_base", "plateau")
                scheduler_afterwards_config = scheduler_config.get("scheduler_afterwards_config", {})
                if scheduler_afterwards_base is not None and isinstance(scheduler_afterwards_base, str):
                    if scheduler_afterwards_base == "plateau":
                        scheduler_afterwards_config = update_dict(scheduler_afterwards_config, plateau_default_config)
                    scheduler_afterwards_base = scheduler_dict[scheduler_afterwards_base]
                scheduler_config["scheduler_afterwards_base"] = scheduler_afterwards_base
                scheduler_config["scheduler_afterwards_config"] = scheduler_afterwards_config
            if scheduler is None:
                self.schedulers[params_name] = None
            else:
                if isinstance(scheduler, str):
                    scheduler = scheduler_dict[scheduler]
                self.schedulers[params_name] = scheduler(opt, **scheduler_config)
        self.schedulers_requires_metric = set()
        for key, scheduler in self.schedulers.items():
            signature = inspect.signature(scheduler.step)
            for name, param in signature.parameters.items():
                if param.kind is inspect.Parameter.POSITIONAL_OR_KEYWORD:
                    if name == "metrics":
                        self.schedulers_requires_metric.add(key)

    def _init_metrics(self):
        # metrics
        metric_config = self.config.setdefault("metric_config", {})
        metric_types = metric_config.setdefault("types", "auto")
        if metric_types == "auto":
            if self.tr_data.is_reg:
                metric_types = ["mae", "mse"]
            else:
                metric_types = ["auc", "acc"]
        if not isinstance(metric_types, (list, tuple)):
            metric_types = [metric_types]
        self._metrics_need_loss = False
        self.metrics: Dict[str, Union[Metrics, None]] = {}
        self.metrics_decay: Dict[str, ScalarEMA] = {}
        metric_decay = metric_config.setdefault("decay", 0.1)
        for i, metric_type in enumerate(metric_types):
            if metric_type not in Metrics.sign_dict:
                # here, sub_metric must be one of the `loss_dict` keys
                self._metrics_need_loss = True
                self.metrics[metric_type] = None
            else:
                sub_metric_config = metric_config.setdefault(f"{metric_type}_config", {})
                self.metrics[metric_type] = Metrics(metric_type, sub_metric_config)
            if not 0. < metric_decay < 1.:
                self.metrics_decay = None
            else:
                self.metrics_decay[metric_type] = ScalarEMA(metric_decay)
        self.metrics_weights = metric_config.setdefault("weights", {})
        for metric_type in metric_types:
            self.metrics_weights.setdefault(metric_type, 1.)

    @property
    def start_snapshot(self):
        return self._step_count >= self.snapshot_start_step and self._epoch_count > self.min_epoch

    @property
    def start_monitor_plateau(self):
        return self._step_count >= self.plateau_monitor_start_snapshot * self.num_step_per_snapshot

    @property
    def num_step_per_snapshot(self):
        if self._num_step_per_snapshot > 0:
            return self._num_step_per_snapshot
        return max(1, min(
            self.max_step_per_snapshot,
            int(len(self.tr_loader) / self.num_snapshot_per_epoch)
        ))

    # core

    def _collect_info(self, *, return_only: bool):
        msg = "\n".join(["=" * 100, "configurations", "-" * 100, ""])
        msg += pprint.pformat(self._wrapper_config, compact=True) + "\n" + "-" * 100 + "\n"
        msg += "\n".join(["=" * 100, "parameters", "-" * 100, ""])
        for name, param in self.model.named_parameters():
            if param.requires_grad:
                msg += name + "\n"
        msg += "\n".join(["-" * 100, "=" * 100, "buffers", "-" * 100, ""])
        for name, param in self.named_buffers():
            msg += name + "\n"
        msg += "\n".join(["-" * 100, "=" * 100, "structure", "-" * 100, str(self), "-" * 100, ""])
        if not return_only:
            self.log_block_msg(msg, verbose_level=4)
        all_msg, msg = msg, "=" * 100 + "\n"
        n_tr = len(self.tr_data)
        n_cv = None if self.cv_data is None else len(self.cv_data)
        msg += f"{self.info_prefix}training data : {n_tr}\n"
        msg += f"{self.info_prefix}valid    data : {n_cv}\n"
        msg += "-" * 100
        if not return_only:
            self.log_block_msg(msg, verbose_level=3)
        return "\n".join([all_msg, msg])

    def _to_device(self, arr: Union[np.ndarray, None]) -> Union[torch.Tensor, None]:
        if arr is None:
            return arr
        return torch.from_numpy(arr).to(self.model.device)

    def _collate_batch(self,
                       x_batch: np.ndarray,
                       y_batch: np.ndarray) -> tensor_dict_type:
        tensors = list(map(self._to_device, [x_batch, y_batch]))
        if y_batch is not None and self.tr_data.is_clf:
            tensors[-1] = tensors[-1].to(torch.int64)
        return dict(zip(["x_batch", "y_batch"], tensors))

    def _clip_norm_step(self):
        self._gradient_norm = torch.nn.utils.clip_grad_norm_(
            self.model.parameters(), self._clip_norm)

    def _optimizer_step(self):
        self.model._optimizer_step(self.optimizers)

    def _monitor_step(self):
        if self._step_count % self.num_step_per_snapshot == 0:
            score, metrics = self._get_metrics()
            if self.start_monitor_plateau:
                if not self._monitor.plateau_flag:
                    self.log_msg("start monitoring plateau", self.info_prefix, 2)
                self._monitor.plateau_flag = True
            if self.start_snapshot:
                if self._monitor.check_terminate(score):
                    return True
                for key, scheduler in self.schedulers.items():
                    if scheduler is not None:
                        kwargs = {}
                        if key in self.schedulers_requires_metric:
                            kwargs["metrics"] = score
                        scheduler.step(**kwargs)
        return False

    def _get_metrics(self) -> Tuple[float, Dict[str, float]]:
        tr_loader, cv_loader = self.tr_loader, self.cv_loader
        if cv_loader is None and self.tr_loader._n_siamese > 1:
            raise ValueError("cv set should be provided when n_siamese > 1")
        if cv_loader is not None:
            loader = cv_loader
        else:
            loader = tr_loader.copy()
            loader.enabled_sampling = False
        if self.use_tqdm:
            loader = tqdm(loader, total=len(loader), leave=False, position=2)
        if not self._metrics_need_loss:
            losses = None
            results = self._predict(loader=loader)
            predictions, labels = map(results.get, ["predictions", "labels"])
        else:
            predictions = None
            forward_dicts, loss_dicts, labels = [], [], []
            for x_batch, y_batch in loader:
                labels.append(y_batch)
                batch = self._collate_batch(x_batch, y_batch)
                with eval_context(self):
                    forward_dicts.append(self.model(batch))
                    loss_dicts.append(self.model.loss_function(batch, forward_dicts[-1]))
            losses, forwards = map(self.model._collate_tensor_dicts, [loss_dicts, forward_dicts])
            losses = {k: v.mean().item() for k, v in losses.items()}
        use_decayed = False
        signs, metrics, decayed_metrics = {}, {}, {}
        for metric_type, metric_ins in self.metrics.items():
            if metric_ins is None:
                signs[metric_type] = -1
                sub_metric = metrics[metric_type] = losses[metric_type]
            else:
                signs[metric_type] = metric_ins.sign
                if self.tr_data.is_reg:
                    metric_predictions = predictions
                else:
                    if metric_ins.requires_prob:
                        metric_predictions = self.model.to_prob(predictions)
                    else:
                        metric_predictions = predictions.argmax(1).reshape([-1, 1])
                sub_metric = metrics[metric_type] = metric_ins.metric(labels, metric_predictions)
            if self.metrics_decay is not None and self.start_snapshot:
                use_decayed = True
                decayed_metrics[metric_type] = self.metrics_decay[metric_type].update("metric", sub_metric)
        metrics_for_scoring = decayed_metrics if use_decayed else metrics
        weighted_scores = [v * signs[k] * self.metrics_weights[k] for k, v in metrics_for_scoring.items()]
        score = sum(weighted_scores) / len(weighted_scores)

        if self._epoch_tqdm is not None:
            self._epoch_tqdm.set_postfix(metrics_for_scoring)

        def _metric_verbose(k):
            metric_str = fix_float_to_length(metrics[k], 8)
            if not use_decayed:
                return metric_str
            return f"{metric_str} (ema: {fix_float_to_length(decayed_metrics[k], 8)})"

        msg = (
            f"| epoch {self._epoch_count:^4d} - step {self._step_count:^6d} | "
            f"{' | '.join([f'{k} : {_metric_verbose(k)}' for k in sorted(metrics)])} | "
            f"score : {fix_float_to_length(score, 8)} |"
        )
        with open(self._log_file, "a") as f:
            f.write(f"{msg}\n")
        self.log_msg(msg, verbose_level=None)

        return score, metrics

    def _predict(self,
                 loader: DataLoader,
                 **kwargs) -> Dict[str, np.ndarray]:
        with eval_context(self, no_grad=kwargs.get("no_grad", False)):
            results, labels = [], []
            for x_batch, y_batch in loader:
                if y_batch is not None:
                    labels.append(y_batch)
                results.append(self.model(self._collate_batch(x_batch, y_batch), **kwargs))
        results = self.model._collate_tensor_dicts(results)
        results = {k: to_numpy(v) for k, v in results.items()}
        if labels:
            labels = np.vstack(labels)
            results["labels"] = labels
        return results

    # api

    def forward(self,
                tr_data: TabularData,
                cv_data: TabularData = None):
        # data
        self._init_data(tr_data, cv_data)
        # optimizer
        self._init_optimizers()
        # metrics
        self._init_metrics()
        # monitor
        self._monitor = TrainMonitor(1).register_pipeline(self)
        # train
        self._collect_info(return_only=False)
        terminate = False
        self._step_count = self._epoch_count = 0
        tuple(map(lambda n: os.makedirs(n, exist_ok=True), [self.logging_folder, self.checkpoint_folder]))
        log_name = f"{timestamp()}.txt"
        self._log_file = os.path.join(self.logging_folder, log_name)
        with open(self._log_file, "w"):
            pass
        self._step_tqdm = self._epoch_tqdm = None
        if self.use_tqdm:
            self._epoch_tqdm = tqdm(list(range(self.num_epoch)), position=0)
        while self._epoch_count < self.num_epoch:
            try:
                self._epoch_count += 1
                self._step_tqdm = iter(self.tr_loader)
                if self.use_tqdm:
                    self._step_tqdm = tqdm(self._step_tqdm, total=len(self.tr_loader), position=1, leave=False)
                for x_batch, y_batch in self._step_tqdm:
                    self._step_count += 1
                    with timing_context(self, "collate batch"):
                        batch = self._collate_batch(x_batch, y_batch)
                    with timing_context(self, "model.forward"):
                        forward_results = self.model(batch)
                    with timing_context(self, "loss.forward"):
                        loss_dict = self.model.loss_function(batch, forward_results)
                    with timing_context(self, "loss.backward"):
                        loss_dict["loss"].backward()
                    if self._clip_norm > 0.:
                        with timing_context(self, "clip_norm_step"):
                            self._clip_norm_step()
                    with timing_context(self, "optimizer_step"):
                        self._optimizer_step()
                    if self.ema_decay is not None:
                        with timing_context(self, "EMA"):
                            self.ema_decay()
                    with timing_context(self, "monitor_step"):
                        terminate = self._monitor_step()
                    if terminate:
                        break
            except KeyboardInterrupt:
                self.log_msg("keyboard interrupted", self.error_prefix, msg_level=logging.ERROR)
                terminate = True
            if terminate:
                if os.path.isdir(self.checkpoint_folder):
                    self.log_msg("rolling back to the best checkpoint", self.info_prefix, 3)
                    self.restore_checkpoint()
                break
            if self.use_tqdm:
                self._epoch_tqdm.total = self.num_epoch
                self._epoch_tqdm.update()
        if self.use_tqdm:
            if self._step_tqdm is not None:
                self._step_tqdm.close()
            self._epoch_tqdm.close()
        self._step_count = self._epoch_count = -1
        self._get_metrics()

    def predict(self,
                x: data_type,
                return_all: bool = False,
                **kwargs) -> Union[np.ndarray, Dict[str, np.ndarray]]:
        data = self.tr_data.copy_to(x, None)
        loader = DataLoader(self.cv_batch_size, ImbalancedSampler(data, shuffle=False))
        predictions = self._predict(loader, **kwargs)
        if return_all:
            return predictions
        return predictions["predictions"]

    @staticmethod
    def _filter_checkpoints(folder) -> Dict[int, str]:
        checkpoints = {}
        for file in os.listdir(folder):
            if file.startswith("pipeline_") and file.endswith(".pt"):
                step = int(os.path.splitext(file)[0].split("_")[1])
                checkpoints[step] = file
        return checkpoints

    def save_checkpoint(self, folder=None):
        if folder is None:
            folder = self.checkpoint_folder
        if self.max_snapshot_num > 0:
            checkpoints = self._filter_checkpoints(folder)
            if len(checkpoints) >= self.max_snapshot_num - 1:
                for key in sorted(checkpoints)[:-self.max_snapshot_num+1]:
                    os.remove(os.path.join(folder, checkpoints[key]))
        file = f"pipeline_{self._step_count}.pt"
        torch.save(self.state_dict(), os.path.join(folder, file))

    def restore_checkpoint(self, folder=None):
        if folder is None:
            folder = self.checkpoint_folder
        checkpoints = self._filter_checkpoints(folder)
        if not checkpoints:
            self.log_msg(
                f"no pipeline file found in {self.checkpoint_folder}",
                self.warning_prefix, msg_level=logging.WARNING
            )
            return self
        latest_checkpoint = checkpoints[sorted(checkpoints)[-1]]
        pipeline_file = os.path.join(folder, latest_checkpoint)
        self.log_msg(f"restoring from {pipeline_file}", self.info_prefix, 4)
        self.load_state_dict(torch.load(pipeline_file, map_location=self.model.device))
        return self


class Wrapper(LoggingMixin):
    def __init__(self,
                 config: Union[str, Dict[str, Any]] = None,
                 *,
                 increment_config: Union[str, Dict[str, Any]] = None,
                 cuda: Union[str, int] = None,
                 verbose_level: int = 2):
        self._verbose_level = int(verbose_level)
        if cuda is not None:
            self.device = torch.device(f"cuda:{cuda}")
        else:
            self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.config, increment_config = map(self._get_config, [config, increment_config])
        update_dict(increment_config, self.config)
        self._init_config()

    @property
    def train_set(self) -> TabularDataset:
        raw = self.tr_data.raw
        return TabularDataset(*raw.xy, task_type=self.tr_data.task_type)

    @property
    def valid_set(self) -> Union[TabularDataset, None]:
        if self.cv_data is None:
            return
        raw = self.cv_data.raw
        return TabularDataset(*raw.xy, task_type=self.cv_data.task_type)

    @property
    def binary_threshold(self) -> Union[float, None]:
        return self._binary_threshold

    @staticmethod
    def _get_config(config):
        if config is None:
            return {}
        if isinstance(config, str):
            with open(config, "r") as f:
                return json.load(f)
        return shallow_copy_dict(config)

    def _init_config(self):
        self._data_config = self.config.setdefault("data_config", {})
        self._data_config["default_categorical_process"] = "identical"
        self._read_config = self.config.setdefault("read_config", {})
        self._cv_split = self.config.setdefault("cv_split", 0.1)
        self._model = self.config.setdefault("model", "fcnn")
        self._binary_metric = self.config.setdefault("binary_metric", "acc")
        self._is_binary = self.config.get("is_binary")
        self._binary_threshold = self.config.get("binary_threshold")
        logging_folder = self.config["logging_folder"] = self.config.setdefault(
            "logging_folder",
            os.path.join("_logging", model_dict[self._model].__identifier__)
        )
        logging_file = self.config.get("logging_file")
        if logging_file is not None:
            logging_path = os.path.join(logging_folder, logging_file)
        else:
            logging_path = os.path.abspath(self.generate_logging_path(logging_folder))
        self.config["_logging_path_"] = logging_path
        self._init_logging(self._verbose_level, self.config.setdefault("trigger_logging", False))

    def _prepare_modules(self):
        # model
        with timing_context(self, "init model"):
            self.model = model_dict[self._model](self.config, self.tr_data, self.device)
        # pipeline
        with timing_context(self, "init pipeline"):
            self.pipeline = Pipeline(self.model, self.config, self._verbose_level)
        # to device
        with timing_context(self, "init device"):
            self.pipeline.to(self.device)

    # api

    def fit(self,
            x: data_type,
            y: data_type = None,
            x_cv: data_type = None,
            y_cv: data_type = None) -> "Wrapper":
        # data
        y, y_cv = map(to_2d, [y, y_cv])
        args = (x, y) if y is not None else (x,)
        self._data_config["verbose_level"] = self._verbose_level
        self.tr_data = TabularData(**self._data_config).read(*args, **self._read_config)
        if x_cv is not None:
            self.cv_data = self.tr_data.copy_to(x_cv, y_cv)
        else:
            if self._cv_split <= 0.:
                self.cv_data = None
            else:
                self.cv_data, self.tr_data = self.tr_data.split(self._cv_split)
        # modules
        self._prepare_modules()
        # training loop
        self.pipeline(self.tr_data, self.cv_data)
        # binary threshold
        if self._binary_threshold is None:
            if self.tr_data.num_classes != 2:
                self._is_binary = False
                self._binary_threshold = None
            else:
                self._is_binary = True
                x, y = self.tr_data.raw.x, self.tr_data.processed.y
                probabilities = self.predict_prob(x)
                try:
                    threshold = Metrics.get_binary_threshold(y, probabilities, self._binary_metric)
                    self._binary_threshold = threshold
                except ValueError:
                    self._binary_threshold = None
        # logging
        self.log_timing()
        return self

    def predict(self,
                x: data_type,
                *,
                return_all: bool = False,
                requires_recover: bool = True,
                **kwargs) -> Union[np.ndarray, Dict[str, np.ndarray]]:
        if self.tr_data.is_reg:
            predictions = self.pipeline.predict(x, return_all, **kwargs)
            recover = partial(self.tr_data.recover_labels, inplace=True)
            if not return_all:
                if requires_recover:
                    return recover(predictions)
                return predictions
            if not requires_recover:
                return predictions
            return {k: recover(v) for k, v in predictions.items()}
        probabilities = self.predict_prob(x, **kwargs)
        if not self._is_binary or self._binary_threshold is None:
            return probabilities.argmax(1).reshape([-1, 1])
        return (probabilities[..., 1] >= self._binary_threshold).astype(np.int).reshape([-1, 1])

    def predict_prob(self,
                     x: data_type,
                     **kwargs) -> np.ndarray:
        if self.tr_data.is_reg:
            raise ValueError("`predict_prob` should not be called on regression tasks")
        raw = self.pipeline.predict(x, **kwargs)
        return self.model.to_prob(raw)

    def save(self,
             folder: str = None,
             *,
             compress: bool = True) -> "Wrapper":
        if folder is None:
            folder = self.pipeline.checkpoint_folder
        abs_folder = os.path.abspath(folder)
        base_folder = os.path.dirname(abs_folder)
        with lock_manager(base_folder, [folder]):
            Saving.prepare_folder(self, folder)
            self.tr_data.save(os.path.join(folder, "__data__", "train"), compress=compress)
            if self.cv_data is not None:
                self.cv_data.save(os.path.join(folder, "__data__", "valid"), compress=compress)
            self.pipeline.save_checkpoint(folder)
            self.config["is_binary"] = self._is_binary
            self.config["binary_threshold"] = self._binary_threshold
            Saving.save_dict(self.config, "config", folder)
            if compress:
                Saving.compress(abs_folder, remove_original=True)
        return self

    @classmethod
    def load(cls,
             folder: str,
             *,
             cuda: int = None,
             verbose_level: int = 0,
             compress: bool = True) -> "Wrapper":
        base_folder = os.path.dirname(os.path.abspath(folder))
        with lock_manager(base_folder, [folder]):
            with Saving.compress_loader(folder, compress, remove_extracted=True):
                config = Saving.load_dict("config", folder)
                wrapper = Wrapper(config, cuda=cuda, verbose_level=verbose_level)
                tr_data_folder = os.path.join(folder, "__data__", "train")
                cv_data_folder = os.path.join(folder, "__data__", "valid")
                tr_data = wrapper.tr_data = TabularData.load(tr_data_folder, compress=compress)
                cv_data = None
                if os.path.isdir(cv_data_folder) or os.path.isfile(f"{cv_data_folder}.zip"):
                    cv_data = wrapper.cv_data = TabularData.load(cv_data_folder, compress=compress)
                wrapper._prepare_modules()
                pipeline = wrapper.pipeline
                pipeline.restore_checkpoint(folder)
                pipeline._init_data(tr_data, cv_data)
                pipeline._init_metrics()
        return wrapper


__all__ = ["ModelBase", "model_dict", "Pipeline", "Wrapper"]



import os

import numpy as np
import matplotlib.pyplot as plt

from cftool.misc import show_or_save


class DDRPredictor:
    def __init__(self, ddr):
        self.m = ddr

    def mr(self, x):
        return self.m.predict(x, predict_median_residual=True, return_all=True)

    def cdf(self, x, y, *, get_pdf: bool = False):
        predictions = self.m.predict(
            x,
            y=y,
            no_grad=not get_pdf,
            requires_recover=False,
            predict_pdf=get_pdf, predict_cdf=True, return_all=True
        )
        if get_pdf:
            return predictions
        return predictions["cdf"]

    def quantile(self, x, q):
        return self.m.predict(x, q=q, predict_quantile=True, return_all=True)["quantile"]


class DDRVisualizer:
    def __init__(self, ddr):
        self.m = ddr
        self.predictor = DDRPredictor(ddr)

    @staticmethod
    def _prepare_base_figure(x, y, x_base, mean, median, indices, title):
        figure = plt.figure()
        plt.title(title)
        plt.scatter(x[indices], y[indices], color="gray", s=15)
        if mean is not None:
            plt.plot(x_base.ravel(), mean.ravel(), label="mean")
        plt.plot(x_base.ravel(), median.ravel(), label="median")
        return figure

    @staticmethod
    def _render_figure(x_min, x_max, y_min, y_max, y_padding):
        plt.xlim(x_min, x_max)
        plt.ylim(y_min - 0.5 * y_padding, y_max + 0.5 * y_padding)
        plt.legend()

    def visualize(self, x, y, export_path,
                  *,
                  residual=False, quantiles=None, anchor_ratios=None, **kwargs):
        x_min, x_max = np.min(x), np.max(x)
        y_min, y_max = np.min(y), np.max(y)
        padding, dense = kwargs.get("padding", 1.), kwargs.get("dense", 400)
        indices = np.random.permutation(len(x))[:dense]
        x_padding = max(abs(x_min), abs(x_max)) * padding
        y_padding = max(abs(y_min), abs(y_max)) * padding
        x_min -= x_padding
        x_max += x_padding
        x_base = np.linspace(x_min, x_max, dense)[..., None]
        mean = None
        median = self.m.predict(x_base)
        fig = DDRVisualizer._prepare_base_figure(x, y, x_base, mean, median, indices, "")
        render_args = x_min, x_max, y_min, y_max, y_padding
        # median residual
        if residual:
            residuals = self.predictor.mr(x_base)
            pos, neg = map(residuals.get, ["mr_pos", "mr_neg"])
            plt.plot(x_base.ravel(), pos, label="pos_median_residual")
            plt.plot(x_base.ravel(), neg, label="neg_median_residual")
            DDRVisualizer._render_figure(*render_args)
        # quantile curves
        if quantiles is not None:
            for q in quantiles:
                quantile_curve = self.predictor.quantile(x_base, q)
                plt.plot(x_base.ravel(), quantile_curve, label=f"quantile {q:4.2f}")
            DDRVisualizer._render_figure(*render_args)
        # cdf curves
        if anchor_ratios is not None:
            y_abs_max = np.abs(y).max()
            ratios, anchors = anchor_ratios, [ratio * (y_max - y_min) + y_min for ratio in anchor_ratios]
            for ratio, anchor in zip(ratios, anchors):
                predictions = self.predictor.cdf(x_base, anchor, get_pdf=True)
                pdf, cdf = map(predictions.get, ["pdf", "cdf"])
                pdf, cdf = pdf * (y_abs_max / max(np.abs(pdf).max(), 1e-8)), cdf * y_abs_max
                anchor_line = np.full(len(x_base), anchor)
                plt.plot(x_base.ravel(), pdf, label=f"pdf {ratio:4.2f}")
                plt.plot(x_base.ravel(), cdf, label=f"cdf {ratio:4.2f}")
                plt.plot(x_base.ravel(), anchor_line, label=f"anchor {ratio:4.2f}", color="gray")
            DDRVisualizer._render_figure(x_min, x_max, y_min, y_max, y_padding)
        show_or_save(export_path, fig)

    def visualize_multiple(self, x, y, x_base, y_matrix, export_folder,
                           *,
                           quantiles=None, anchor_ratios=None):
        y_min, y_max = y.min(), y.max()
        y_diff = y_max - y_min

        def _plot(prefix, num, y_true, predictions, dual_predictions):
            def _core(pred, *, dual):
                suffix = "_dual" if dual else ""
                plt.figure()
                plt.title(f"{prefix} {num:6.4f}")
                plt.scatter(x[:200], y[:200], color="gray", s=15)
                plt.plot(x_base, y_true, label="target")
                plt.plot(x_base, pred, label=f"ddr{suffix}_prediction")
                plt.legend()
                show_or_save(os.path.join(export_folder, f"{prefix}_{num:4.2f}{suffix}.png"))

            _core(predictions, dual=False)
            if dual_predictions is not None:
                _core(dual_predictions, dual=True)

        if quantiles is not None:
            for quantile in quantiles:
                yq = np.percentile(y_matrix, int(100 * quantile), axis=1)
                yq_pred = self.predictor.quantile(x_base, quantile)
                # yqd_pred = ddr.quantile(quantile, x_base, inference_only=True)
                yqd_pred = None
                _plot("quantile", quantile, yq, yq_pred, yqd_pred)
            plt.figure()
            for quantile in [0.25, 0.5, 0.75]:
                yq = np.percentile(y_matrix, int(100 * quantile), axis=1)
                plt.title(f"quantile {quantile:4.2f}")
                plt.scatter(x[:200], y[:200], color="gray", s=15)
                plt.plot(x_base, yq, label=f"{quantile:3.2f}")
                plt.legend()
            export_path = os.path.join(export_folder, "median_residual.png")
            self.visualize(x, y, export_path, residual=True)
        if anchor_ratios is not None:
            anchors = [ratio * (y_max - y_min) + y_min for ratio in anchor_ratios]
            for anchor in anchors:
                yd = np.mean(y_matrix <= anchor, axis=1) * y_diff + y_min
                yd_pred = self.predictor.cdf(x_base, anchor)
                yd_pred = yd_pred * y_diff + y_min
                # ydq_pred = ddr.cdf(data_anchor, x_base, inference_only=True) * y_diff + y_min
                ydq_pred = None
                _plot("cdf", anchor, yd, yd_pred, ydq_pred)


__all__ = ["DDRPredictor", "DDRVisualizer"]



import torch

import numpy as np
import torch.nn as nn

from typing import *
from cfdata.tabular import TabularData
from sklearn.tree import _tree, DecisionTreeClassifier

from ...bases import ModelBase
from ...misc.toolkit import *
from ...modules.blocks import *


def export_structure(tree):
    tree = tree.tree_

    def recurse(node, depth):
        feature_dim = tree.feature[node]
        if feature_dim == _tree.TREE_UNDEFINED:
            yield depth, -1, tree.value[node]
        else:
            threshold = tree.threshold[node]
            yield depth, feature_dim, threshold
            yield from recurse(tree.children_left[node], depth + 1)
            yield depth, feature_dim, threshold
            yield from recurse(tree.children_right[node], depth + 1)

    return tuple(recurse(0, 0))


@ModelBase.register("ndt")
class NDT(ModelBase):
    def __init__(self,
                 config: Dict[str, Any],
                 tr_data: TabularData,
                 device: torch.device):
        super().__init__(config, tr_data, device)
        # prepare
        x, y = tr_data.processed.xy
        y_ravel, num_classes = y.ravel(), tr_data.num_classes
        x_tensor = torch.from_numpy(x).to(device)
        split_result = self._split_features(x_tensor)
        # decision tree
        self.log_msg("fitting decision tree", self.info_prefix, verbose_level=2)
        x_merge = split_result.merge().cpu().numpy()
        self.dt = DecisionTreeClassifier(**self.dt_config, random_state=142857).fit(x_merge, y_ravel)
        tree_structure = export_structure(self.dt)
        # dt statistics
        num_leafs = sum([1 if pair[1] == -1 else 0 for pair in tree_structure])
        num_internals = num_leafs - 1
        msg = f"internals : {num_internals} ; leafs : {num_leafs}"
        self.log_msg(msg, self.info_prefix, verbose_level=2)
        # transform
        b = np.zeros(num_internals, dtype=np.float32)
        w1 = np.zeros([self.merged_dim, num_internals], dtype=np.float32)
        w2 = np.zeros([num_internals, num_leafs], dtype=np.float32)
        w3 = np.zeros([num_leafs, num_classes], dtype=np.float32)
        node_list, node_sign_list = [], []
        node_id_cursor = leaf_id_cursor = 0
        for depth, feat_dim, rs in tree_structure:
            if feat_dim != -1:
                if depth == len(node_list):
                    node_sign_list.append(-1)
                    node_list.append(node_id_cursor)
                    w1[feat_dim, node_id_cursor] = 1
                    b[node_id_cursor] = -rs
                    node_id_cursor += 1
                else:
                    node_list = node_list[:depth + 1]
                    node_sign_list = node_sign_list[:depth] + [1]
            else:
                for node_id, node_sign in zip(node_list, node_sign_list):
                    w2[node_id, leaf_id_cursor] = node_sign / len(node_list)
                w3[leaf_id_cursor] = rs / np.sum(rs)
                leaf_id_cursor += 1
        w1, w2, w3, b = map(torch.from_numpy, [w1, w2, w3, b])
        # construct planes & routes
        self.to_planes = Linear(self.merged_dim, num_internals, init_method=None)
        self.to_routes = Linear(num_internals, num_leafs, bias=False, init_method=None)
        self.to_leafs = Linear(num_leafs, num_classes, init_method=None)
        with torch.no_grad():
            self.to_planes.linear.bias.data = b
            self.to_planes.linear.weight.data = w1.t()
            self.to_routes.linear.weight.data = w2.t()
            self.to_leafs.linear.weight.data = w3.t()
            uniform = nn.functional.log_softmax(torch.zeros(num_classes, dtype=torch.float32), dim=0)
            self.to_leafs.linear.bias.data = uniform

    @property
    def hyperplane_weights(self) -> np.ndarray:
        return to_numpy(self.to_planes.linear.weight)

    @property
    def hyperplane_thresholds(self) -> np.ndarray:
        return to_numpy(-self.to_planes.linear.bias)

    @property
    def route_weights(self) -> np.ndarray:
        return to_numpy(self.to_routes.linear.weight)

    @property
    def class_log_distributions(self) -> np.ndarray:
        return to_numpy(self.to_leafs.linear.weight)

    @property
    def class_log_prior(self) -> np.ndarray:
        return to_numpy(self.to_leafs.linear.bias)

    @property
    def class_prior(self) -> np.ndarray:
        return np.exp(self.class_log_prior)

    def _preset_config(self,
                       tr_data: TabularData):
        self.config.setdefault("default_encoding_method", "one_hot")

    def _init_config(self,
                     tr_data: TabularData):
        super()._init_config(tr_data)
        self.dt_config = self.config.setdefault("dt_config", {})
        activation_configs = self.config.setdefault("activation_configs", {})
        activation_configs.setdefault("multiplied_tanh", {}).setdefault("ratio", 10.)
        activation_configs.setdefault("multiplied_softmax", {}).setdefault("ratio", 10.)
        default_activations = {"planes": "sign", "routes": "multiplied_softmax"}
        activations = self.config.setdefault("activations", default_activations)
        activations_ins = Activations(activation_configs)
        self.planes_activation = activations_ins.module(activations.get("planes"))
        self.routes_activation = activations_ins.module(activations.get("routes"))
        self._init_with_dt = self.config.setdefault("")

    def forward(self,
                batch: tensor_dict_type,
                **kwargs) -> tensor_dict_type:
        x_batch = batch["x_batch"]
        merged = self._split_features(x_batch).merge()
        planes = self.planes_activation(self.to_planes(merged))
        routes = self.routes_activation(self.to_routes(planes))
        leafs = self.to_leafs(routes)
        return {"predictions": leafs}


__all__ = ["NDT"]



import torch
import torch.nn as nn

from typing import *
from functools import partial
from torch.nn.functional import softplus

from ..misc.toolkit import tensor_dict_type


class BN(nn.BatchNorm1d):
    def forward(self, net):
        if len(net.shape) == 3:
            net = net.transpose(1, 2)
        net = super().forward(net)
        if len(net.shape) == 3:
            net = net.transpose(1, 2)
        return net


class Dropout(nn.Module):
    def __init__(self, dropout: float):
        if dropout < 0. or dropout > 1.:
            raise ValueError(f"dropout probability has to be between 0 and 1, but got {dropout}")
        super().__init__()
        self._mask_cache = None
        self._keep_prob = 1. - dropout

    def forward(self, net, *, reuse: bool = False):
        if not self.training:
            return net
        if reuse:
            mask = self._mask_cache
        else:
            self._mask_cache = mask = torch.bernoulli(
                net.new(*net.shape).fill_(self._keep_prob)) / self._keep_prob
        net = net * mask
        del mask
        return net

    def extra_repr(self) -> str:
        return f"keep={self._keep_prob}"


class EMA(nn.Module):
    def __init__(self,
                 decay: float,
                 named_parameters: Iterable[Tuple[str, nn.Parameter]]):
        super().__init__()
        self._decay, self._named_parameters = decay, list(named_parameters)
        for name, param in self.tgt_params:
            self.register_buffer("tr_" + name, param.data.clone())
            self.register_buffer("ema_" + name, param.data.clone())

    @property
    def tgt_params(self):
        return map(
            lambda pair: (pair[0].replace(".", "_"), pair[1]),
            filter(lambda pair: pair[1].requires_grad, self._named_parameters)
        )

    def forward(self):
        for name, param in self.tgt_params:
            tr_name, ema_name = "tr_" + name, "ema_" + name
            setattr(self, tr_name, param.data.clone())
            ema = (1.0 - self._decay) * param.data + self._decay * getattr(self, ema_name)
            setattr(self, ema_name, ema.clone())

    def train(self, mode: bool = True):
        super().train(mode)
        prefix = "tr_" if mode else "ema_"
        for name, param in self.tgt_params:
            param.data = getattr(self, prefix + name).clone()

    def extra_repr(self) -> str:
        max_str_len = max(len(name) for name, _ in self.tgt_params)
        return "\n".join(
            [f"(0): decay_rate={self._decay}\n(1): Params("] + [
                f"  {name:<{max_str_len}s} - torch.Tensor({list(param.shape)})"
                for name, param in self.tgt_params
            ] + [")"]
        )


class MTL(nn.Module):
    def __init__(self,
                 num_tasks: int,
                 method: str = None):
        super().__init__()
        self._n_task, self._method = num_tasks, method
        if method is None or method == "naive":
            pass
        elif method == "softmax":
            self.w = torch.nn.Parameter(torch.ones(num_tasks))
        else:
            raise NotImplementedError(f"MTL method '{method}' not implemented")
        self._slice, self.registered, self._registered = None, False, {}

    def register(self, names: Iterable[str]) -> None:
        if self.registered:
            raise ValueError("re-register is not permitted")
        self._rev_registered = {}
        for name in sorted(names):
            idx = len(self._registered)
            self._registered[name], self._rev_registered[idx] = idx, name
        self._slice, self.registered = len(names), True
        if self._slice > self._n_task:
            raise ValueError("registered names are more than n_task")

    def forward(self,
                loss_dict: tensor_dict_type,
                naive: bool = False) -> torch.Tensor:
        if not self.registered:
            raise ValueError("losses need to be registered")
        if naive or self._method is None:
            return self._naive(loss_dict)
        return getattr(self, f"_{self._method}")(loss_dict)

    @staticmethod
    def _naive(loss_dict: tensor_dict_type) -> torch.Tensor:
        return sum(loss_dict.values())

    def _softmax(self, loss_dict: tensor_dict_type) -> torch.Tensor:
        w = self.w if self._slice == self._n_task else self.w[:self._slice]
        softmax_w = nn.functional.softmax(w, dim=0)
        losses = []
        for key, loss in loss_dict.items():
            idx = self._registered.get(key)
            losses.append(loss if idx is None else loss * softmax_w[idx])
        return sum(losses) * self._slice

    def extra_repr(self) -> str:
        method = "naive" if self._method is None else self._method
        return f"n_task={self._n_task}, method='{method}'"


class Pruner(nn.Module):
    def __init__(self, config: Dict[str, Any]):
        super().__init__()
        tensor = partial(torch.tensor, dtype=torch.float32)
        self.method = config.setdefault("method", "auto_prune")
        if self.method == "surgery":
            self.register_buffer("alpha", tensor([config.setdefault("alpha", 1.)]))
            self.register_buffer("beta", tensor([config.setdefault("beta", 4.)]))
            self.register_buffer("gamma", tensor([config.setdefault("gamma", 1e-4)]))
            self.register_buffer("eps", tensor([config.setdefault("eps", 1e-12)]))
            keys = ["alpha", "beta", "gamma", "eps"]
            self._mask = None
        elif self.method == "simplified":
            self.register_buffer("alpha", tensor([config.setdefault("alpha", 0.01)]))
            self.register_buffer("beta", tensor([config.setdefault("beta", 1.)]))
            self.register_buffer("max_ratio", tensor([config.setdefault("max_ratio", 1.)]))
            self.register_buffer("exp", tensor([config.setdefault("exp", 0.5)]))
            keys = ["alpha", "beta", "max_ratio", "exp"]
        else:
            self.register_buffer("alpha", tensor([
                config.setdefault("alpha", 1e-4 if self.method == "hard_prune" else 1e-2)]))
            self.register_buffer("beta", tensor([config.setdefault("beta", 1.)]))
            self.register_buffer("gamma", tensor([config.setdefault("gamma", 1.)]))
            self.register_buffer("max_ratio", tensor([config.setdefault("max_ratio", 1.)]))
            if not all(scalar > 0 for scalar in [self.alpha, self.beta, self.gamma, self.max_ratio]):
                raise ValueError("parameters should greater than 0. in pruner")
            self.register_buffer("eps", tensor([config.setdefault("eps", 1e-12)]))
            if self.method == "auto_prune":
                for attr in ["alpha", "beta", "gamma", "max_ratio"]:
                    setattr(self, attr, torch.log(torch.exp(getattr(self, attr)) - 1))
                self.alpha, self.beta, self.gamma, self.max_ratio = map(
                    lambda param: nn.Parameter(param),
                    [self.alpha, self.beta, self.gamma, self.max_ratio]
                )
            keys = ["alpha", "beta", "gamma", "max_ratio", "eps"]
        self._repr_keys = keys
        self.device = None

    def forward(self, w, prune: bool = True):
        if self.device is None:
            self.device = w.device
        if not prune:
            return w
        w_abs = torch.abs(w)
        if self.method == "surgery":
            if self._mask is None:
                self._mask = torch.ones_like(w, dtype=torch.float32).to(self.device)
            mu, std = torch.mean(w_abs), torch.std(w_abs)
            ones_mask, zeros_mask = self._mask.eq(1.), self._mask.eq(0.)
            to_zeros_mask = ones_mask & (w_abs <= 0.9 * (mu - self.beta * std))
            to_ones_mask = zeros_mask & (w_abs >= 1.1 * (mu + self.beta * std))
            self._mask[to_zeros_mask], self._mask[to_ones_mask] = 0., 1.
            mask = self._mask
            del mu, std, ones_mask, zeros_mask, to_zeros_mask, to_ones_mask
        else:
            if self.method != "auto_prune":
                alpha, beta, ratio = self.alpha, self.beta, self.max_ratio
            else:
                alpha, beta, ratio = map(softplus, [self.alpha, self.beta, self.max_ratio])
            if self.method == "simplified":
                log_w = torch.min(ratio, beta * w_abs ** self.exp)
            else:
                w_abs_mean = torch.mean(w_abs)
                gamma = self.gamma if self.method != "auto_prune" else softplus(self.gamma)
                log_w = torch.log(torch.max(self.eps, w_abs / (w_abs_mean * gamma)))
                log_w = torch.min(ratio, beta * log_w)
                del w_abs_mean
            mask = torch.max(alpha / beta * log_w, log_w)
            del log_w
        w = w * mask
        del w_abs, mask
        return w

    def extra_repr(self) -> str:
        if self.method == "auto_prune":
            return f"method='{self.method}'"
        max_str_len = max(map(len, self._repr_keys))
        return "\n".join(
            [f"(0): method={self.method}\n(1): Settings("] + [
                f"  {key:<{max_str_len}s} - {getattr(self, key).item()}"
                for key in self._repr_keys
            ] + [")"]
        )


__all__ = ["BN", "Dropout", "EMA", "MTL", "Pruner"]



import os
import json
import argparse

import numpy as np

import cflearn


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("--config_file")
    args = parser.parse_args()
    with open(args.config_file, "r") as f:
        config = json.load(f)
    logging_folder = config["logging_folder"]
    keys = ["x", "y", "x_cv", "y_cv"]
    data_list = list(map(config.pop, keys, 4 * [None]))
    if data_list[0] is None:
        for i, key in enumerate(keys):
            data_file = os.path.join(logging_folder, f"{key}.npy")
            if os.path.isfile(data_file):
                data_list[i] = np.load(data_file)
    m = cflearn.make(**config)
    m.fit(*data_list)
    cflearn.save(m, saving_folder=logging_folder)



import torch
import logging

import torch.nn as nn

from typing import *
from cftool.ml import Anneal
from cftool.misc import LoggingMixin

from ...misc.toolkit import tensor_dict_type
from ...modules.auxiliary import MTL


class DDRLoss(nn.Module, LoggingMixin):
    def __init__(self,
                 config: Dict[str, Any],
                 device: torch.device):
        super().__init__()
        self._joint_training = config["joint_training"]
        self._use_dynamic_dual_loss_weights = config["use_dynamic_weights"]
        self._use_anneal, self._anneal_step = config["use_anneal"], config["anneal_step"]
        self._median_pressure = config.setdefault("median_pressure", 3.)
        self._median_pressure_inv = 1. / self._median_pressure
        self.mtl = MTL(16, config["mtl_method"])
        self._target_loss_warned = False
        self._zero = torch.zeros([1], dtype=torch.float32).to(device)
        if self._use_anneal:
            anneal_config = config.setdefault("anneal_config", {})
            anneal_methods = anneal_config.setdefault("methods", {})
            anneal_ratios = anneal_config.setdefault("ratios", {})
            anneal_floors = anneal_config.setdefault("floors", {})
            anneal_ceilings = anneal_config.setdefault("ceilings", {})
            default_anneal_methods = {
                "median_anneal": "linear", "main_anneal": "linear",
                "monotonous_anneal": "sigmoid", "anchor_anneal": "linear",
                "dual_anneal": "sigmoid", "recover_anneal": "sigmoid", "pressure_anneal": "sigmoid"
            }
            default_anneal_ratios = {
                "median_anneal": 0.25, "main_anneal": 0.25,
                "monotonous_anneal": 0.2, "anchor_anneal": 0.2,
                "dual_anneal": 0.75, "recover_anneal": 0.75, "pressure_anneal": 0.5
            }
            default_anneal_floors = {
                "median_anneal": 1., "main_anneal": 0.,
                "monotonous_anneal": 0., "anchor_anneal": 0.,
                "dual_anneal": 0., "recover_anneal": 0., "pressure_anneal": 0.
            }
            default_anneal_ceilings = {
                "median_anneal": 2.5, "main_anneal": 0.8,
                "monotonous_anneal": 2.5, "anchor_anneal": 2.,
                "dual_anneal": 0.1, "recover_anneal": 0.1, "pressure_anneal": 1.,
            }
            for anneal in default_anneal_methods:
                anneal_methods.setdefault(anneal, default_anneal_methods[anneal])
                anneal_ratios.setdefault(anneal, default_anneal_ratios[anneal])
                anneal_floors.setdefault(anneal, default_anneal_floors[anneal])
                anneal_ceilings.setdefault(anneal, default_anneal_ceilings[anneal])
            for anneal in default_anneal_methods:
                attr = f"_{anneal}"
                if anneal_methods[anneal] is None:
                    setattr(self, attr, None)
                else:
                    setattr(self, attr, Anneal(
                        anneal_methods[anneal], round(self._anneal_step * anneal_ratios[anneal]),
                        anneal_floors[anneal], anneal_ceilings[anneal]
                    ))

    def forward(self,
                predictions: tensor_dict_type,
                target: torch.Tensor,
                *,
                check_monotonous_only: bool = False) -> Tuple[torch.Tensor, tensor_dict_type]:
        # anneal
        if not self._use_anneal or not self.training or check_monotonous_only:
            main_anneal = median_anneal = None
            monotonous_anneal = anchor_anneal = None
            dual_anneal = recover_anneal = pressure_anneal = None
        else:
            main_anneal = None if self._main_anneal is None else self._main_anneal.pop()
            median_anneal = None if self._median_anneal is None else self._median_anneal.pop()
            monotonous_anneal = None if self._monotonous_anneal is None else self._monotonous_anneal.pop()
            anchor_anneal = None if self._median_anneal is None else self._anchor_anneal.pop()
            dual_anneal = None if self._median_anneal is None else self._dual_anneal.pop()
            recover_anneal = None if self._median_anneal is None else self._recover_anneal.pop()
            pressure_anneal = None if self._pressure_anneal is None else self._pressure_anneal.pop()
            self._last_main_anneal, self._last_pressure_anneal = main_anneal, pressure_anneal
        if self._use_anneal and check_monotonous_only:
            main_anneal, pressure_anneal = self._last_main_anneal, self._last_pressure_anneal
        # median
        median = predictions["predictions"]
        median_loss = nn.functional.l1_loss(median, target)
        if median_anneal is not None:
            median_loss = median_loss * median_anneal
        # get
        anchor_batch, cdf_raw = map(predictions.get, ["anchor_batch", "cdf_raw"])
        sampled_anchors, sampled_cdf_raw = map(predictions.get, ["sampled_anchors", "sampled_cdf_raw"])
        quantile_batch, median_residual, quantile_residual, quantile_sign = map(
            predictions.get, ["quantile_batch", "median_residual", "quantile_residual", "quantile_sign"])
        sampled_quantiles, sampled_quantile_residual = map(
            predictions.get, ["sampled_quantiles", "sampled_quantile_residual"])
        cdf_gradient, quantile_residual_gradient = map(
            predictions.get, ["cdf_gradient", "quantile_residual_gradient"])
        dual_quantile, quantile_cdf_raw = map(predictions.get, ["dual_quantile", "quantile_cdf_raw"])
        dual_cdf, cdf_quantile_residual = map(predictions.get, ["dual_cdf", "cdf_quantile_residual"])
        # cdf
        fetch_cdf = cdf_raw is not None
        cdf_anchor_loss = cdf_monotonous_loss = None
        if not fetch_cdf or check_monotonous_only:
            cdf_loss = cdf_losses = None
        else:
            cdf_losses = self._get_cdf_loss(target, cdf_raw, anchor_batch, False)
            if main_anneal is not None:
                cdf_losses = cdf_losses * main_anneal
            cdf_loss = cdf_losses.mean()
            if sampled_cdf_raw is not None:
                cdf_anchor_loss = self._get_cdf_loss(target, sampled_cdf_raw, sampled_anchors, True)
                if anchor_anneal is not None:
                    cdf_anchor_loss = cdf_anchor_loss * anchor_anneal
        # cdf monotonous
        if cdf_gradient is not None:
            cdf_monotonous_loss = nn.functional.relu(-cdf_gradient).mean()
            if anchor_anneal is not None:
                cdf_monotonous_loss = cdf_monotonous_loss * monotonous_anneal
        # quantile
        fetch_quantile = quantile_residual is not None
        quantile_anchor_loss = quantile_monotonous_loss = None
        if not fetch_quantile or check_monotonous_only:
            median_residual_loss = quantile_loss = quantile_losses = None
        else:
            target_median_residual = target - predictions["median_detach"]
            median_residual_loss = self._get_median_residual_loss(
                target_median_residual, median_residual, quantile_sign)
            if anchor_anneal is not None:
                median_residual_loss = median_residual_loss * anchor_anneal
            quantile_losses = self._get_quantile_residual_loss(
                target_median_residual, quantile_residual, quantile_batch, False)
            quantile_loss = quantile_losses.mean() + median_residual_loss
            if main_anneal is not None:
                quantile_loss = quantile_loss * main_anneal
            if sampled_quantile_residual is not None:
                quantile_anchor_loss = self._get_quantile_residual_loss(
                    target_median_residual, sampled_quantile_residual,
                    sampled_quantiles, True
                )
                if anchor_anneal is not None:
                    quantile_anchor_loss = quantile_anchor_loss * anchor_anneal
        # median pressure
        if not fetch_quantile:
            median_pressure_loss = None
        else:
            median_pressure_loss = self._get_median_pressure_loss(predictions)
            if pressure_anneal is not None:
                median_pressure_loss = median_pressure_loss * pressure_anneal
        # quantile monotonous
        quantile_monotonous_losses = []
        if quantile_residual_gradient is not None:
            quantile_monotonous_losses.append(nn.functional.relu(-quantile_residual_gradient).mean())
        if median_residual is not None and quantile_sign is not None:
            quantile_monotonous_losses.append(
                self._get_median_residual_monotonous_loss(median_residual, quantile_sign))
        if quantile_monotonous_losses:
            quantile_monotonous_loss = sum(quantile_monotonous_losses)
            if anchor_anneal is not None:
                quantile_monotonous_loss = quantile_monotonous_loss * monotonous_anneal
        # dual
        if not self._joint_training or not fetch_cdf or not fetch_quantile or check_monotonous_only:
            dual_cdf_loss = dual_quantile_loss = None
            cdf_recover_loss = quantile_recover_loss = None
        else:
            # dual cdf (cdf -> quantile [recover loss] -> cdf [dual loss])
            quantile_recover_loss, quantile_recover_losses, quantile_recover_loss_weights = \
                self._get_dual_recover_loss(dual_quantile, anchor_batch, cdf_losses)
            if quantile_cdf_raw is None:
                dual_quantile_loss = None
            else:
                dual_quantile_losses = self._get_cdf_loss(target, quantile_cdf_raw, anchor_batch, False)
                if quantile_recover_losses is None or not self._use_dynamic_dual_loss_weights:
                    dual_quantile_loss_weights = 1.
                else:
                    quantile_recover_losses_detach = quantile_recover_losses.detach()
                    dual_quantile_loss_weights = 0.5 * (
                        quantile_recover_loss_weights + 1 / (1 + 2 * torch.tanh(quantile_recover_losses_detach)))
                dual_quantile_loss = (dual_quantile_losses * dual_quantile_loss_weights).mean()
            # dual quantile (quantile -> cdf [recover loss] -> quantile [dual loss])
            cdf_recover_loss, cdf_recover_losses, cdf_recover_loss_weights = \
                self._get_dual_recover_loss(dual_cdf, quantile_batch, quantile_losses)
            if cdf_quantile_residual is None:
                dual_cdf_loss = None
            else:
                dual_cdf_losses = self._get_quantile_residual_loss(
                    target, cdf_quantile_residual, quantile_batch, False)
                if cdf_recover_losses is None or not self._use_dynamic_dual_loss_weights:
                    dual_cdf_loss_weights = 1.
                else:
                    cdf_recover_losses_detach = cdf_recover_losses.detach()
                    dual_cdf_loss_weights = 0.5 * (
                        cdf_recover_loss_weights + 1 / (1 + 10 * cdf_recover_losses_detach))
                dual_cdf_loss = (dual_cdf_losses * dual_cdf_loss_weights).mean() + median_residual_loss
        if dual_anneal is not None:
            if dual_cdf_loss is not None:
                dual_cdf_loss = dual_cdf_loss * dual_anneal
            if dual_quantile_loss is not None:
                dual_quantile_loss = dual_quantile_loss * dual_anneal
        if recover_anneal is not None:
            if cdf_recover_loss is not None:
                cdf_recover_loss = cdf_recover_loss * recover_anneal
            if quantile_recover_loss is not None:
                quantile_recover_loss = quantile_recover_loss * recover_anneal
        # combine
        if check_monotonous_only:
            losses = {}
        else:
            losses = {"median": median_loss}
            if not self._joint_training:
                if cdf_anchor_loss is not None:
                    losses["cdf_anchor"] = cdf_anchor_loss
                if quantile_anchor_loss is not None:
                    losses["quantile_anchor"] = quantile_anchor_loss
            else:
                if fetch_cdf:
                    losses["cdf"] = cdf_loss
                    if cdf_anchor_loss is not None:
                        losses["cdf_anchor"] = cdf_anchor_loss
                if fetch_quantile:
                    losses["quantile"] = quantile_loss
                    if quantile_anchor_loss is not None:
                        losses["quantile_anchor"] = quantile_anchor_loss
                if fetch_cdf and fetch_quantile:
                    losses["quantile_recover"], losses["cdf_recover"] = quantile_recover_loss, cdf_recover_loss
                    losses["dual_quantile"], losses["dual_cdf"] = dual_quantile_loss, dual_cdf_loss
        if median_residual_loss is not None:
            losses["median_residual_loss"] = median_residual_loss
        if median_pressure_loss is not None:
            key = "synthetic_median_pressure_loss" if check_monotonous_only else "median_pressure_loss"
            losses[key] = median_pressure_loss
        if cdf_monotonous_loss is not None:
            key = "synthetic_cdf_monotonous" if check_monotonous_only else "cdf_monotonous"
            losses[key] = cdf_monotonous_loss
        if quantile_monotonous_loss is not None:
            key = "synthetic_quantile_monotonous" if check_monotonous_only else "quantile_monotonous"
            losses[key] = quantile_monotonous_loss
        if not losses:
            return self._zero, {"loss": self._zero}
        if not self.mtl.registered:
            self.mtl.register(losses.keys())
        return self.mtl(losses), losses

    def _get_dual_recover_loss(self, dual_prediction, another_input_batch, another_losses):
        if dual_prediction is None:
            recover_loss = recover_losses = recover_loss_weights = None
        else:
            recover_losses = torch.abs(another_input_batch - dual_prediction)
            if not self._use_dynamic_dual_loss_weights:
                recover_loss_weights = 1.
            else:
                another_losses_detach = another_losses.detach()
                recover_loss_weights = 1 / (1 + 2 * torch.tanh(another_losses_detach))
            recover_loss = (recover_losses * recover_loss_weights).mean()
        return recover_loss, recover_losses, recover_loss_weights

    @staticmethod
    def _get_cdf_loss(target, cdf_raw, anchor_batch, reduce):
        indicative = (target <= anchor_batch).to(torch.float32)
        cdf_losses = -indicative * cdf_raw + nn.functional.softplus(cdf_raw)
        return cdf_losses if not reduce else cdf_losses.mean()

    @staticmethod
    def _get_median_residual_monotonous_loss(median_residual, quantile_sign):
        return nn.functional.relu(-median_residual * quantile_sign).mean()

    @staticmethod
    def _get_quantile_residual_loss(target_residual, quantile_residual, quantile_batch, reduce):
        quantile_error = target_residual - quantile_residual
        quantile_losses = torch.max(quantile_batch * quantile_error, (quantile_batch - 1) * quantile_error)
        return quantile_losses if not reduce else quantile_losses.mean()

    def _get_median_residual_loss(self, target_median_residual, median_residual, quantile_sign):
        same_sign_mask = quantile_sign * torch.sign(target_median_residual) > 0
        tmr, mr = map(lambda tensor: tensor[same_sign_mask], [target_median_residual, median_residual])
        median_residual_mae = self._median_pressure * torch.abs(tmr - mr).mean()
        residual_monotonous_loss = DDRLoss._get_median_residual_monotonous_loss(median_residual, quantile_sign)
        return median_residual_mae + residual_monotonous_loss

    def _get_median_pressure_loss(self, predictions):
        pressure_pos_dict, pressure_neg_dict = map(
            predictions.get, map(lambda attr: f"pressure_sub_quantile_{attr}_dict", ["pos", "neg"]))
        additive_pos, additive_neg = pressure_pos_dict["add"], pressure_neg_dict["add"]
        multiply_pos, multiply_neg = pressure_pos_dict["mul"], pressure_neg_dict["mul"]
        # additive net & multiply net are tend to be zero here
        # because median pressure batch receives 0.5 as input
        return sum(
            torch.max(
                -self._median_pressure * sub_quantile,
                self._median_pressure_inv * sub_quantile
            ).mean()
            for sub_quantile in [
                additive_pos, -additive_neg,
                multiply_pos, multiply_neg
            ]
        )


__all__ = ["DDRLoss"]



import torch

import numpy as np
import torch.nn as nn

from typing import *

from .auxiliary import *
from ..misc.toolkit import *


class Linear(nn.Module):
    def __init__(self,
                 in_dim: int,
                 out_dim: int,
                 *,
                 bias: bool = True,
                 pruner_config: dict = None,
                 init_method: Union[str, None] = "xavier",
                 **kwargs):
        super().__init__()
        self.linear = nn.Linear(in_dim, out_dim, bias)
        pruner = None if pruner_config is None else Pruner(pruner_config)
        self.config, self.pruner = kwargs, pruner
        self._use_bias, self._init_method = bias, init_method
        with torch.no_grad():
            self.reset_parameters()

    @property
    def weight(self) -> torch.Tensor:
        return self.linear.weight

    @property
    def bias(self) -> Union[torch.Tensor, None]:
        return self.linear.bias

    def forward(self, net):
        weight = self.linear.weight if self.pruner is None else self.pruner(self.linear.weight)
        return nn.functional.linear(net, weight, self.linear.bias)

    def reset_parameters(self):
        if self._init_method is None:
            return
        if self._init_method not in Initializer.defined_initialization:
            return
        initializer = Initializer(self.config.setdefault("initialize_config", {}))
        initializer.initialize(self.linear.weight, self._init_method)
        bias_fill = self.config.setdefault("bias_fill", 0.)
        if self._use_bias:
            self.linear.bias.data.fill_(bias_fill)


class Mapping(nn.Module):
    def __init__(self,
                 in_dim: int,
                 out_dim: int,
                 *,
                 bias: bool = True,
                 pruner_config: dict = None,
                 dropout: bool = True,
                 batch_norm: bool = True,
                 activation: str = "ReLU",
                 init_method: str = "xavier",
                 **kwargs):
        super().__init__()
        self.config = kwargs
        self.linear = Linear(
            in_dim, out_dim, bias=bias,
            pruner_config=pruner_config, init_method=init_method, **kwargs
        )
        self.bn = None if not batch_norm else BN(out_dim)
        if activation is None:
            self.activation = None
        else:
            activations_ins = Activations(self.config.setdefault("activation_config", None))
            self.activation = activations_ins.module(activation)
        self.dropout = None if not dropout else Dropout(self.config.setdefault("drop_prob", 0.5))

    @property
    def weight(self) -> torch.Tensor:
        return self.linear.weight

    @property
    def bias(self) -> Union[torch.Tensor, None]:
        return self.linear.bias

    def forward(self, net, *, reuse: bool = False):
        net = self.linear(net)
        if self.bn is not None:
            net = self.bn(net)
        if self.activation is not None:
            net = self.activation(net)
        if self.dropout is not None:
            net = self.dropout(net, reuse=reuse)
        return net


class MLP(nn.Module):
    def __init__(self,
                 in_dim: int,
                 out_dim: Union[int, None],
                 num_units: List[int],
                 mapping_configs: List[Dict[str, Any]],
                 *,
                 final_mapping_config: Dict[str, Any] = None):
        super().__init__()
        mappings = []
        for num_unit, mapping_config in zip(num_units, mapping_configs):
            mappings.append(Mapping(in_dim, num_unit, **mapping_config))
            in_dim = num_unit
        if out_dim is not None:
            if final_mapping_config is None:
                final_mapping_config = {}
            mappings.append(Linear(in_dim, out_dim, **final_mapping_config))
        self.mappings = nn.ModuleList(mappings)

    @property
    def weights(self) -> List[torch.Tensor]:
        return [mapping.weight for mapping in self.mappings]

    @property
    def biases(self) -> List[Union[torch.Tensor, None]]:
        return [mapping.bias for mapping in self.mappings]

    def forward(self, net):
        for mapping in self.mappings:
            net = mapping(net)
        return net


class DNDF(nn.Module):
    def __init__(self,
                 in_dim: int,
                 out_dim: int,
                 *,
                 num_tree: int = 10,
                 tree_depth: int = 4,
                 tree_proj_config: dict = None,
                 output_type: str = "output",
                 is_regression: bool = None):
        super().__init__()
        self._num_tree, self._tree_depth, self._output_type = num_tree, tree_depth, output_type
        self._is_regression = out_dim == 1 if is_regression is None else is_regression
        self._num_leaf = 2 ** (self._tree_depth + 1)
        self._num_internals = self._num_leaf - 1
        self._output_dim = out_dim
        tree_proj_config = self._setup_tree_proj(tree_proj_config)
        self.tree_proj = Linear(in_dim, self._num_internals * self._num_tree, **tree_proj_config)
        self.leafs = nn.Parameter(torch.empty(self._num_tree, self._num_leaf, self._output_dim))

    @staticmethod
    def _setup_tree_proj(tree_proj_config):
        if tree_proj_config is None:
            tree_proj_config = {}
        pruner_config = tree_proj_config.pop("pruner_config", {})
        if pruner_config is not None:
            tree_proj_config["pruner"] = Pruner(pruner_config)
        return tree_proj_config

    def forward(self, net):
        device = net.device
        tree_net = self.tree_proj(net)
        p_left = torch.split(torch.sigmoid(tree_net), self._num_internals, dim=-1)
        flat_probabilities = [torch.reshape(torch.cat([p, 1. - p], dim=-1), [-1]) for p in p_left]
        n_flat_prob = 2 * self._num_internals
        batch_indices = torch.reshape(torch.arange(
            0, n_flat_prob * net.shape[0], n_flat_prob), [-1, 1]).to(device)
        num_repeat, num_local_internals = self._num_leaf // 2, 1
        increment_mask = torch.from_numpy(np.repeat(
            [0, self._num_internals], num_repeat).astype(np.int64)).to(device)
        routes = [p_flat.take(batch_indices + increment_mask) for p_flat in flat_probabilities]
        for _ in range(1, self._tree_depth + 1):
            num_repeat //= 2
            num_local_internals *= 2
            increment_mask = np.repeat(np.arange(num_local_internals - 1, 2 * num_local_internals - 1), 2)
            increment_mask += np.tile([0, self._num_internals], num_local_internals)
            increment_mask = np.repeat(increment_mask, num_repeat)
            increment_mask = torch.from_numpy(increment_mask.astype(np.int64)).to(device)
            for i, p_flat in enumerate(flat_probabilities):
                routes[i] *= p_flat.take(batch_indices + increment_mask)
        leafs, features = self.leafs, torch.cat(routes, 1)
        if not self._is_regression and self._output_dim > 1:
            leafs = nn.functional.softmax(leafs, dim=-1)
        leafs = leafs.view(self._num_tree * self._num_leaf, self._output_dim)
        return features.matmul(leafs) / self._num_tree

    def reset_parameters(self):
        self.tree_proj.reset_parameters()
        nn.init.xavier_uniform_(self.leafs.data)


__all__ = ["Linear", "Mapping", "MLP", "DNDF"]



import torch
import logging

from typing import *
from cfdata.tabular import TabularData

from ..fcnn.core import FCNN
from ...bases import SplitFeatures
from ...misc.toolkit import tensor_dict_type
from ...modules.blocks import *


@FCNN.register("tree_dnn")
class TreeDNN(FCNN):
    def __init__(self,
                 config: Dict[str, Any],
                 tr_data: TabularData,
                 device: torch.device):
        super(FCNN, self).__init__(config, tr_data, device)
        encoding_dims = self.encoding_dims
        # fc
        self._use_embedding_for_fc = self.config.setdefault("use_embedding_for_fc", True)
        self._use_one_hot_for_fc = self.config.setdefault("use_one_hot_for_fc", False)
        if not self._numerical_columns:
            self._use_embedding_for_fc = self._use_one_hot_for_fc = True
        fc_in_dim = self.merged_dim
        if not self._use_embedding_for_fc:
            fc_in_dim -= encoding_dims.get("embedding", 0)
        if not self._use_one_hot_for_fc:
            fc_in_dim -= encoding_dims.get("one_hot", 0)
        self._fc_in_dim = fc_in_dim
        self._init_fcnn()
        # dndf
        dndf_config = config.setdefault("dndf_config", {})
        if dndf_config is None:
            self.log_msg(
                "DNDF is not used in TreeDNN, it will be equivalent to FCNN",
                self.warning_prefix, 2, logging.WARNING
            )
            self.dndf = None
        else:
            dndf_config["is_regression"] = tr_data.is_reg
            dndf_config.setdefault("tree_proj_config", None)
            self._use_embedding_for_dndf = self.config.setdefault("use_embedding_for_dndf", True)
            self._use_one_hot_for_dndf = self.config.setdefault("use_one_hot_for_dndf", True)
            if not self._numerical_columns:
                self._use_one_hot_for_dndf = self._use_embedding_for_dndf = True
            dndf_input_dim = self.merged_dim
            if not self._use_embedding_for_dndf:
                dndf_input_dim -= encoding_dims["embedding"]
            if not self._use_one_hot_for_dndf:
                dndf_input_dim -= encoding_dims["one_hot"]
            self.dndf = DNDF(dndf_input_dim, self._fc_out_dim, **dndf_config)

    def _preset_config(self,
                       tr_data: TabularData):
        self.config.setdefault("default_encoding_method", ["one_hot", "embedding"])

    @staticmethod
    def _merge(split_result: SplitFeatures,
               use_embedding: bool,
               use_one_hot: bool) -> torch.Tensor:
        if use_embedding and use_one_hot:
            return split_result.merge()
        numerical = split_result.numerical
        if not use_embedding and not use_one_hot:
            return numerical
        categorical = split_result.categorical
        if not categorical:
            return numerical
        if not use_one_hot:
            return torch.cat([numerical, categorical["embedding"]], dim=1)
        if not use_embedding:
            return torch.cat([numerical, categorical["one_hot"]], dim=1)

    def forward(self,
                batch: tensor_dict_type,
                **kwargs) -> tensor_dict_type:
        x_batch = batch["x_batch"]
        split_result = self._split_features(x_batch, return_all_encodings=True)
        # fc
        fc_net = self._merge(split_result, self._use_embedding_for_fc, self._use_one_hot_for_fc)
        fc_net = self.mlp(fc_net)
        # dndf
        dndf_net = self._merge(split_result, self._use_embedding_for_dndf, self._use_one_hot_for_dndf)
        dndf_net = self.dndf(dndf_net)
        return {"predictions": fc_net + dndf_net}


__all__ = ["TreeDNN"]



import os
import json
import torch
import platform
import subprocess

import numpy as np

from ..misc.toolkit import data_type

IS_LINUX = platform.system() == "Linux"


class Task:
    def __init__(self,
                 idx: int,
                 model: str,
                 identifier: str,
                 temp_folder: str):
        self.idx = idx
        self.model = model
        self.identifier = identifier
        self.temp_folder = temp_folder

    def __str__(self):
        return f"Task({self.identifier}_{self.idx})"

    __repr__ = __str__

    @property
    def saving_folder(self) -> str:
        folder = os.path.join(self.temp_folder, self.identifier, str(self.idx))
        folder = os.path.abspath(folder)
        os.makedirs(folder, exist_ok=True)
        return folder

    @property
    def cuda_run_command(self) -> str:
        python = subprocess.check_output(["which", "python"]).decode().strip()
        return f"{python} -m {'.'.join(['cflearn', 'dist', 'cuda_run'])}"

    def fit(self,
            make: callable,
            save: callable,
            x: data_type,
            y: data_type = None,
            x_cv: data_type = None,
            y_cv: data_type = None,
            cuda: int = None,
            **kwargs) -> "Task":
        kwargs["cuda"] = cuda
        kwargs["logging_folder"] = self.saving_folder
        if not IS_LINUX or not torch.cuda.is_available():
            m = make(self.model, **kwargs)
            m.fit(x, y, x_cv, y_cv)
            save(m, saving_folder=self.saving_folder)
        else:
            kwargs["model"] = self.model
            kwargs["trigger_logging"] = True
            config_file = os.path.join(self.saving_folder, "config.json")
            if not isinstance(x, np.ndarray):
                kwargs["x"], kwargs["y"] = x, y
                kwargs["x_cv"], kwargs["y_cv"] = x_cv, y_cv
            else:
                for key, value in zip(["x", "y", "x_cv", "y_cv"], [x, y, x_cv, y_cv]):
                    if value is None:
                        continue
                    np.save(key, value)
            with open(config_file, "w") as f:
                json.dump(kwargs, f)
            os.system(f"{self.cuda_run_command} --config_file {config_file}")
        return self


__all__ = ["Task"]



import os
import torch

from typing import *
from itertools import product
from cftool.dist import Parallel

from .task import Task
from ..misc.toolkit import data_type


class Experiments:
    def __init__(self,
                 available_cuda_list: List[int] = None):
        if available_cuda_list is None and not torch.cuda.is_available():
            available_cuda_list = []
        self.cuda_list = available_cuda_list

    def run(self,
            make: callable,
            save: callable,
            load_task: callable,
            x: data_type,
            y: data_type = None,
            x_cv: data_type = None,
            y_cv: data_type = None,
            *,
            num_repeat: int = 5,
            num_parallel: int = 4,
            temp_folder: str = "__tmp__",
            models: Union[str, List[str]] = "fcnn",
            identifiers: Union[str, List[str]] = None,
            return_tasks: bool = False,
            use_tqdm_in_task: bool = False,
            use_tqdm: bool = True,
            **kwargs) -> Dict[str, List[Any]]:
        if isinstance(models, str):
            models = [models]
        if identifiers is None:
            identifiers = models.copy()
        elif isinstance(identifiers, str):
            identifiers = [identifiers]
        kwargs["use_tqdm"] = use_tqdm_in_task

        def _task(i, id_tuple, cuda=None):
            t = Task(i, *id_tuple, temp_folder)
            return t.fit(make, save, x, y, x_cv, y_cv, cuda, **kwargs)

        arguments = list(product(map(str, range(num_repeat)), zip(models, identifiers)))
        parallel_arguments = list(zip(*arguments))
        parallel = Parallel(
            num_parallel,
            use_tqdm=use_tqdm,
            use_cuda=torch.cuda.is_available(),
            logging_folder=os.path.join(temp_folder, "_parallel_"),
            resource_config={"gpu_config": {"available_cuda_list": self.cuda_list}}
        )
        tasks = parallel(_task, *parallel_arguments).ordered_results
        results = {}
        for task, (_, (_, identifier)) in zip(tasks, arguments):
            if return_tasks:
                results.setdefault(identifier, []).append(task)
            else:
                loaded = load_task(task)
                results.setdefault(identifier, []).append(loaded)
        return results


__all__ = ["Experiments"]



import math
import torch
import logging

import numpy as np
import torch.nn as nn

from typing import *
from cftool.misc import *
from abc import ABCMeta, abstractmethod

tensor_dict_type = Dict[str, torch.Tensor]
data_type = Union[np.ndarray, List[List[float]], List[float], str, None]


def to_torch(arr: np.ndarray) -> torch.Tensor:
    return torch.from_numpy(arr.astype(np.float32))


def to_numpy(tensor: torch.Tensor) -> np.ndarray:
    return tensor.detach().cpu().numpy()


def to_2d(arr: data_type) -> data_type:
    if arr is None or isinstance(arr, str):
        return
    if isinstance(arr, np.ndarray):
        return arr.reshape([len(arr), -1])
    if isinstance(arr[0], list):
        return arr
    return [[elem] for elem in arr]


def get_gradient(y, x, retain_graph=False, create_graph=False):
    grads = torch.autograd.grad(y, x, torch.ones_like(y), retain_graph, create_graph)
    if len(grads) == 1:
        return grads[0]
    return grads


class Initializer(LoggingMixin):
    """
    Initializer for neural network weights

    Examples
    --------
    >>> initializer = Initializer({})
    >>> linear = nn.Linear(10, 10)
    >>> initializer.xavier(linear.weight)

    """

    defined_initialization = {"xavier", "normal", "truncated_normal"}
    custom_initializer = {}

    def __init__(self, config):
        self.config = config
        self._verbose_level = config.setdefault("verbose_level", 2)

    def initialize(self, param: nn.Parameter, method: str):
        custom_initializer = self.custom_initializer.get(method)
        if custom_initializer is None:
            return getattr(self, method)(param)
        return custom_initializer(self, param)

    @classmethod
    def add_initializer(cls, f, name):
        if name in cls.defined_initialization:
            print(f"{cls.warning_prefix}'{name}' initializer is already defined")
            return
        cls.defined_initialization.add(name)
        cls.custom_initializer[name] = f

    @staticmethod
    def xavier(param: nn.Parameter):
        nn.init.xavier_uniform_(param.data)

    def normal(self, param: nn.Parameter):
        mean = self.config.setdefault("mean", 0.)
        std = self.config.setdefault("std", 1.)
        with torch.no_grad():
            param.data.normal_(mean, std)

    def truncated_normal(self, param: nn.Parameter):
        span = self.config.setdefault("span", 2.)
        mean = self.config.setdefault("mean", 0.)
        std = self.config.setdefault("std", 1.)
        tol = self.config.setdefault("tol", 0.)
        epoch = self.config.setdefault("epoch", 20)
        n_elem = param.numel()
        weight_base = param.new_empty(n_elem).normal_()
        get_invalid = lambda w: (w > span) | (w < -span)
        invalid = get_invalid(weight_base)
        success = False
        for _ in range(epoch):
            n_invalid = int(invalid.sum())
            if n_invalid / n_elem <= tol:
                success = True
                break
            with torch.no_grad():
                weight_base[invalid] = param.new_empty(n_invalid).normal_()
                invalid = get_invalid(weight_base)
        if not success:
            self.log_msg(
                f"invalid ratio for truncated normal : {invalid.to(torch.float32).mean():8.6f}, "
                f"it might cause by too little epoch ({epoch}) or too small tolerance ({tol})",
                self.warning_prefix, 2, logging.WARNING
            )
        with torch.no_grad():
            param.data.copy_(weight_base.reshape(param.shape))
            param.data.mul_(std).add_(mean)


class _multiplied_activation(nn.Module, metaclass=ABCMeta):
    def __init__(self,
                 ratio: float,
                 trainable: bool = True):
        super().__init__()
        self.trainable = trainable
        ratio = torch.tensor([ratio], dtype=torch.float32)
        self.ratio = ratio if not trainable else nn.Parameter(ratio)

    @abstractmethod
    def _core(self, multiplied: torch.Tensor) -> torch.Tensor:
        pass

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self._core(x * self.ratio)

    def extra_repr(self) -> str:
        return f"ratio={self.ratio.item()}, trainable={self.trainable}"


class Activations:
    """
    Wrapper class for pytorch activations
    * when pytorch implemented corresponding activation, it will be returned
    * otherwise, custom implementation will be returned

    Parameters
    ----------
    configs : {None, dict}, configuration for the activation

    Examples
    --------
    >>> act = Activations()
    >>> print(type(act.ReLU))  # <class 'nn.modules.activation.ReLU'>
    >>> print(type(act.module("ReLU")))  # <class 'nn.modules.activation.ReLU'>
    >>> print(type(act.Tanh))  # <class 'nn.modules.activation.Tanh'>
    >>> print(type(act.one_hot))  # <class '__main__.Activations.one_hot.<locals>.OneHot'>

    """

    def __init__(self,
                 configs: Dict[str, Any] = None):
        if configs is None:
            configs = {}
        self.configs = configs

    def __getattr__(self, item):
        try:
            return getattr(nn, item)(**self.configs.setdefault(item, {}))
        except AttributeError:
            raise NotImplementedError(
                f"neither pytorch nor custom Activations implemented activation '{item}'")

    def module(self,
               name: str) -> nn.Module:
        if name is None:
            return nn.Identity()
        return getattr(self, name)

    # publications

    @property
    def mish(self):

        class Mish(nn.Module):
            def forward(self, x):
                return x * (torch.tanh(nn.functional.softplus(x)))

        return Mish()

    # custom

    @property
    def sign(self):

        class Sign(nn.Module):
            def forward(self, x):
                return torch.sign(x)

        return Sign()

    @property
    def one_hot(self):

        class OneHot(nn.Module):
            def forward(self, x):
                return x * (x == torch.max(x, dim=1, keepdim=True)[0]).to(torch.float32)

        return OneHot()

    @property
    def multiplied_sine(self):

        class MultipliedSine(_multiplied_activation):
            def _core(self, multiplied: torch.Tensor) -> torch.Tensor:
                return torch.sin(multiplied)

        config = self.configs.setdefault("multiplied_sine", {})
        config.setdefault("ratio", 10.)
        return MultipliedSine(**config)

    @property
    def multiplied_tanh(self):

        class MultipliedTanh(_multiplied_activation):
            def _core(self, multiplied: torch.Tensor) -> torch.Tensor:
                return torch.tanh(multiplied)

        return MultipliedTanh(**self.configs.setdefault("multiplied_tanh", {}))

    @property
    def multiplied_softmax(self):

        class MultipliedSoftmax(_multiplied_activation):
            def __init__(self, ratio, dim=1, trainable=True):
                super().__init__(ratio, trainable)
                self.dim = dim

            def _core(self, multiplied: torch.Tensor) -> torch.Tensor:
                return nn.functional.softmax(multiplied, dim=self.dim)

        return MultipliedSoftmax(**self.configs.setdefault("multiplied_softmax", {}))


class TrainMonitor:
    """
    Util class to monitor training process of a neural network
    * If overfitting, it will tell the model to early-stop
    * If underfitting, it will tell the model to extend training process
    * If better performance acquired, it will tell the model to save a checkpoint
    * If performance sticks on a plateau, it will tell the model to stop training (to save time)

    Warnings
    ----------
    * Performance MUST be evaluated on the cross validation dataset instead of the training set if possible
    * `register_pipeline` method MUST be called before monitoring
    * instance passed to`register_pipeline` method MUST be a subclass of `LoggingMixin`,
      and must include `_epoch_count`, `num_epoch`, `max_epoch` attributes and
      `save_checkpoint` method

    Parameters
    ----------
    sign : {-1, 1}
        * -1 : the metric will be the lower the better (e.g. mae, mse)
        * 1  : the metric will be the higher the better (e.g. auc, acc)
    num_scores_per_snapshot : int, indicates snapshot frequency
        * `TrainMonitor` will perform a snapshot every `num_scores_per_snapshot` scores are recorded
    history_ratio : float, indicates the ratio of the history's window width
        * history window width will be `num_scores_per_snapshot` * `history_ratio`
    tolerance_ratio : float, indicates the ratio of tolerance
        * tolerance base will be `num_scores_per_snapshot` * `tolerance_ratio`
        * judgements of 'overfitting' and 'performance sticks on a plateau' will based on 'tolerance base'
    extension : int, indicates how much epoch to extend when underfitting occurs
    std_floor : float, indicates the floor of history's std used for judgements
    std_ceiling : float, indicates the ceiling of history's std used for judgements
    aggressive : bool, indicates the strategy of monitoring
        * True  : it will tell the model to save every checkpoint when better metric is reached
        * False : it will be more careful since better metric may lead to
        more seriously over-fitting on cross validation set

    Examples
    ----------
    >>> from cftool.ml import Metrics
    >>>
    >>> x, y, model = ...
    >>> metric = Metrics("mae")
    >>> monitor = TrainMonitor(metric.sign).register_pipeline(model)
    >>> n_epoch, epoch_count = 20, 0
    >>> while epoch_count <= n_epoch:
    >>>     model.train()
    >>>     predictions = model.predict(x)
    >>>     if monitor.check_terminate(metric.metric(y, predictions)):
    >>>         break

    """

    def __init__(self,
                 sign,
                 num_scores_per_snapshot=1,
                 history_ratio=3,
                 tolerance_ratio=2,
                 extension=5,
                 std_floor=0.001,
                 std_ceiling=0.01,
                 aggressive=False):
        self.sign = sign
        self.num_scores_per_snapshot = num_scores_per_snapshot
        self.num_history = int(num_scores_per_snapshot * history_ratio)
        self.num_tolerance = int(num_scores_per_snapshot * tolerance_ratio)
        self.extension = extension
        self.is_aggressive = aggressive
        self.std_floor, self.std_ceiling = std_floor, std_ceiling
        self._scores = []
        self.plateau_flag = False
        self._is_best = self._running_best = None
        self._descend_increment = self.num_history * extension / 30
        self._incrementer = Incrementer(self.num_history)

        self._over_fit_performance = math.inf
        self._best_checkpoint_performance = -math.inf
        self._descend_counter = self._plateau_counter = self.over_fitting_flag = 0
        self.info = {"terminate": False, "save_checkpoint": False, "save_best": aggressive, "info": None}

    @property
    def log_msg(self):
        return self._pipeline.log_msg

    @property
    def plateau_threshold(self):
        return 6 * self.num_tolerance * self.num_history

    def _update_running_info(self, last_score):
        self._incrementer.update(last_score)
        if self._running_best is None:
            if self._scores[0] > self._scores[1]:
                improvement = 0
                self._running_best, self._is_best = self._scores[0], False
            else:
                improvement = self._scores[1] - self._scores[0]
                self._running_best, self._is_best = self._scores[1], True
        elif self._running_best > last_score:
            improvement = 0
            self._is_best = False
        else:
            improvement = last_score - self._running_best
            self._running_best = last_score
            self._is_best = True
        return improvement

    def _handle_overfitting(self, last_score, res, std):
        if self._descend_counter == 0:
            self.info["save_best"] = True
            self._over_fit_performance = last_score
        self._descend_counter += min(self.num_tolerance / 3, -res / std)
        self.log_msg(
            f"descend counter updated : {self._descend_counter:6.4f}",
            self._pipeline.info_prefix, 6, logging.DEBUG
        )
        self.over_fitting_flag = 1

    def _handle_recovering(self, improvement, last_score, res, std):
        if res > 3 * std and self._is_best and improvement > std:
            self.info["save_best"] = True
        new_counter = self._descend_counter - res / std
        if self._descend_counter > 0 >= new_counter:
            self._over_fit_performance = math.inf
            if last_score > self._best_checkpoint_performance:
                self._best_checkpoint_performance = last_score
                if last_score > self._running_best - std:
                    self._plateau_counter //= 2
                    self.info["save_checkpoint"] = True
                    self.info["info"] = (
                        f"current snapshot ({len(self._scores)}) seems to be working well, "
                        "saving checkpoint in case we need to restore"
                    )
            self.over_fitting_flag = 0
        if self._descend_counter > 0:
            self._descend_counter = max(new_counter, 0)
            self.log_msg(
                f"descend counter updated : {self._descend_counter:6.4f}",
                self._pipeline.info_prefix, 6, logging.DEBUG
            )

    def _handle_is_best(self):
        if self._is_best:
            self.info["terminate"] = False
            if self.info["save_best"]:
                self._plateau_counter //= 2
                self.info["save_checkpoint"] = True
                self.info["save_best"] = self.is_aggressive
                self.info["info"] = (
                    f"current snapshot ({len(self._scores)}) leads to best result we've ever had, "
                    "saving checkpoint since "
                )
                if self.over_fitting_flag:
                    self.info["info"] += "we've suffered from over-fitting"
                else:
                    self.info["info"] += "performance has improved significantly"

    def _handle_period(self, last_score):
        if self.is_aggressive:
            return
        if len(self._scores) % self.num_scores_per_snapshot == 0 and \
                last_score > self._best_checkpoint_performance:
            self._best_checkpoint_performance = last_score
            self._plateau_counter //= 2
            self.info["terminate"] = False
            self.info["save_checkpoint"] = True
            self.info["info"] = (
                f"current snapshot ({len(self._scores)}) leads to best checkpoint we've ever had, "
                "saving checkpoint in case we need to restore"
            )

    def _punish_extension(self):
        self.plateau_flag = True
        self._descend_counter += self._descend_increment

    def _handle_pipeline_terminate(self):
        pipeline = self._pipeline
        if self.info["terminate"]:
            self.log_msg(
                f"early stopped at n_epoch={pipeline._epoch_count} due to '{self.info['info']}'",
                prefix=pipeline.info_prefix
            )
            return True
        if self.info["save_checkpoint"]:
            self.log_msg(f"{self.info['info']}", pipeline.info_prefix, 3)
            pipeline.save_checkpoint()
        if (
            pipeline._epoch_count == pipeline.num_epoch and
            pipeline._epoch_count < pipeline.max_epoch and
            not self.info["terminate"]
        ):
            self._punish_extension()
            new_epoch = pipeline.num_epoch + self.extension
            pipeline.num_epoch = min(new_epoch, pipeline.max_epoch)
            self.log_msg(f"extending num_epoch to {pipeline.num_epoch}", pipeline.info_prefix, 3)
        if pipeline._epoch_count == pipeline.max_epoch:
            if not self.info["terminate"]:
                self.log_msg(
                    "model seems to be under-fitting but max_epoch reached, "
                    "increasing max_epoch may improve performance.", pipeline.info_prefix
                )
            return True
        return False

    def register_pipeline(self, pipeline):
        self._pipeline = pipeline
        return self

    def check_terminate(self, new_metric):
        last_score = new_metric * self.sign
        self._scores.append(last_score)
        n_history = min(self.num_history, len(self._scores))
        if math.isnan(new_metric):
            self.info["terminate"] = True
            self.info["info"] = "nan metric encountered"
        elif n_history != 1:
            improvement = self._update_running_info(last_score)
            self.info["save_checkpoint"] = False
            mean, std = self._incrementer.mean, self._incrementer.std
            std = min(std, self.std_ceiling)
            plateau_updated = False
            if std < self.std_floor:
                if self.plateau_flag:
                    self._plateau_counter += self.std_floor / max(std, self.std_floor / 6)
                    plateau_updated = True
            else:
                if self._plateau_counter > 0:
                    self._plateau_counter = max(self._plateau_counter - 1, 0)
                    plateau_updated = True
                res = last_score - mean
                if res < -std and last_score < self._over_fit_performance - std:
                    self._handle_overfitting(last_score, res, std)
                elif res > std:
                    self._handle_recovering(improvement, last_score, res, std)
            if plateau_updated:
                self.log_msg(
                    f"plateau counter updated : {self._plateau_counter:>6.4f} / {self.plateau_threshold}",
                    self._pipeline.info_prefix, 6, logging.DEBUG
                )
            if self._plateau_counter >= self.plateau_threshold:
                self.info["info"] = "performance not improving"
                self.info["terminate"] = True
            else:
                if self._descend_counter >= self.num_tolerance:
                    self.info["info"] = "over-fitting"
                    self.info["terminate"] = True
                else:
                    self._handle_is_best()
                    self._handle_period(last_score)
                    if self.info["save_checkpoint"]:
                        self.info["info"] += " (plateau counter cleared)"
                        self._plateau_counter = 0
        return self._handle_pipeline_terminate()


class eval_context(context_error_handler):
    """
    Help entering eval mode and recovering previous mode

    This is a context controller for entering eval mode at the beginning
    and back to previous mode at the end.

    Useful when we need to predict something with our PyTorch model during training.

    Parameters
    ----------
    module : nn.Module, arbitrary PyTorch module.

    Examples
    --------
    >>> module = nn.Module()
    >>> with eval_context(module):
    >>>     pass  # do something

    """

    def __init__(self,
                 module: nn.Module,
                 *,
                 no_grad: bool = True):
        self._module, self._training = module, module.training
        self._params_required_grad = [param for param in module.parameters() if param.requires_grad]
        tuple(map(lambda param: param.requires_grad_(False), self._params_required_grad))
        self._no_grad = torch.no_grad() if no_grad else None

    def __enter__(self):
        self._module.eval()
        if self._no_grad is not None:
            self._no_grad.__enter__()

    def _normal_exit(self, exc_type, exc_val, exc_tb):
        self._module.train(mode=self._training)
        if self._no_grad is not None:
            self._no_grad.__exit__(exc_type, exc_val, exc_tb)
        tuple(map(lambda param: param.requires_grad_(True), self._params_required_grad))


__all__ = [
    "tensor_dict_type", "data_type",
    "to_torch", "to_numpy", "to_2d", "get_gradient",
    "Initializer", "Activations", "TrainMonitor", "eval_context"
]



import torch

from typing import *
from cfdata.tabular import TabularData

from ...bases import ModelBase
from ...misc.toolkit import tensor_dict_type
from ...modules.blocks import *


@ModelBase.register("fcnn")
class FCNN(ModelBase):
    def __init__(self,
                 config: Dict[str, Any],
                 tr_data: TabularData,
                 device: torch.device):
        super().__init__(config, tr_data, device)
        self._init_fcnn()

    def _init_config(self,
                     tr_data: TabularData):
        super()._init_config(tr_data)
        self._fc_in_dim, self._fc_out_dim = map(self.config.get, ["fc_in_dim", "fc_out_dim"])
        self.hidden_units = self.config.setdefault("hidden_units", [256, 256])
        self.mapping_configs = self.config.setdefault("mapping_configs", {})
        if isinstance(self.mapping_configs, dict):
            self.mapping_configs = [self.mapping_configs] * len(self.hidden_units)
        self.out_dim = max(tr_data.num_classes, 1)

    def _init_fcnn(self):
        if self._fc_in_dim is not None:
            in_dim = self._fc_in_dim
        else:
            in_dim = self._fc_in_dim = self.merged_dim
        if self._fc_out_dim is not None:
            out_dim = self._fc_out_dim
        else:
            out_dim = self._fc_out_dim = self.out_dim
        final_mapping_config = self.config.setdefault("final_mapping_config", {})
        self.mlp = MLP(
            in_dim, out_dim, self.hidden_units, self.mapping_configs,
            final_mapping_config=final_mapping_config
        )

    def forward(self,
                batch: tensor_dict_type,
                **kwargs) -> tensor_dict_type:
        x_batch = batch["x_batch"]
        net = self._split_features(x_batch).merge()
        net = self.mlp(net)
        return {"predictions": net}


__all__ = ["FCNN"]



from setuptools import setup, find_packages

VERSION = "0.1.0"
DESCRIPTION = "A minimal Automatic Machine Learning (AutoML) solution for tabular datasets based on PyTorch"
with open("README.md") as f:
    LONG_DESCRIPTION = f.read()

setup(
    name="carefree-learn",
    version="0.1.0",
    packages=find_packages(exclude=("tests",)),
    install_requires=[
        "carefree-ml>=0.1.0",
        "carefree-data>=0.1.3",
        "carefree-toolkit>=0.1.6",
        "dill", "future", "psutil", "tb-nightly",
        "cython>=0.29.12", "numpy>=1.16.2", "scipy>=1.2.1",
        "scikit-learn>=0.23.1", "matplotlib>=3.0.3",
        "mkdocs", "mkdocs-material", "mkdocs-minify-plugin",
        "Pygments", "pymdown-extensions"
    ],
    author="carefree0910",
    author_email="syameimaru_kurumi@pku.edu.cn",
    url="https://github.com/carefree0910/carefree-learn",
    download_url=f"https://github.com/carefree0910/carefree-learn/archive/v{VERSION}.tar.gz",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    keywords="python automl machine-learning solution PyTorch"
)
